# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime

import mock
from sqlalchemy import exc as sa_exc
import testtools

from sahara import context
from sahara.db.sqlalchemy import models as m
from sahara import exceptions as ex
from sahara.service.castellan import config as castellan
import sahara.tests.unit.conductor.base as test_base
from sahara.tests.unit.conductor.manager import test_clusters
from sahara.utils import edp


SAMPLE_DATA_SOURCE = {
    "tenant_id": "tenant_1",
    "name": "ngt_test",
    "description": "test_desc",
    "type": "Cassandra",
    "url": "localhost:1080",
    "credentials": {
        "user": "test",
        "password": "123"
    },
    "is_public": False,
    "is_protected": False
}

SAMPLE_JOB = {
    "tenant_id": "tenant_1",
    "name": "job_test",
    "description": "test_desc",
    "type": edp.JOB_TYPE_PIG,
    "mains": [],
    "is_public": False,
    "is_protected": False
}

SAMPLE_JOB_EXECUTION = {
    "tenant_id": "tenant_1",
    "return_code": "1",
    "job_id": "undefined",
    "input_id": "undefined",
    "output_id": "undefined",
    "start_time": datetime.datetime.now(),
    "cluster_id": None,
    "is_public": False,
    "is_protected": False
}

SAMPLE_CONF_JOB_EXECUTION = {
    "tenant_id": "tenant_1",
    "progress": "0.1",
    "return_code": "1",
    "job_id": "undefined",
    "input_id": "undefined",
    "output_id": "undefined",
    "cluster_id": None,
    "job_configs": {
        "conf2": "value_je",
        "conf3": "value_je"
    }
}

BINARY_DATA = b"vU}\x97\x1c\xdf\xa686\x08\xf2\tf\x0b\xb1}"

SAMPLE_JOB_BINARY_INTERNAL = {
    "tenant_id": "test_tenant",
    "name": "job_test",
    "data": BINARY_DATA,
    "is_public": False,
    "is_protected": False
}


SAMPLE_JOB_BINARY = {
    "tenant_id": "test_tenant",
    "name": "job_binary_test",
    "description": "test_dec",
    "url": "internal-db://test_binary",
    "is_public": False,
    "is_protected": False
}

SAMPLE_JOB_BINARY_UPDATE = {
    "name": "updatedName",
    "url": "internal-db://updated-fake-url"
}

SAMPLE_JOB_BINARY_SWIFT = {
    "tenant_id": "test_tenant",
    "name": "job_binary_test_swift",
    "description": "the description",
    "url": "swift://test_swift_url",
}

SAMPLE_JOB_BINARY_SWIFT_UPDATE = {
    "name": "SwifterName",
    "url": "swift://updated-swift"
}


class DataSourceTest(test_base.ConductorManagerTestCase):
    def __init__(self, *args, **kwargs):
        super(DataSourceTest, self).__init__(
            checks=[
                lambda: SAMPLE_DATA_SOURCE
            ], *args, **kwargs)

    def setUp(self):
        super(DataSourceTest, self).setUp()
        castellan.validate_config()

    def test_crud_operation_create_list_delete(self):
        ctx = context.ctx()
        self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)

        lst = self.api.data_source_get_all(ctx)
        self.assertEqual(1, len(lst))

        ds_id = lst[0]['id']
        self.api.data_source_destroy(ctx, ds_id)

        lst = self.api.data_source_get_all(ctx)
        self.assertEqual(0, len(lst))

    def test_duplicate_data_source_create(self):
        ctx = context.ctx()
        self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)
        with testtools.ExpectedException(ex.DBDuplicateEntry):
            self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)

    def test_data_source_fields(self):
        ctx = context.ctx()
        ctx.tenant_id = SAMPLE_DATA_SOURCE['tenant_id']
        ds_db_obj_id = self.api.data_source_create(ctx,
                                                   SAMPLE_DATA_SOURCE)['id']

        ds_db_obj = self.api.data_source_get(ctx, ds_db_obj_id)
        self.assertIsInstance(ds_db_obj, dict)

        for key, val in SAMPLE_DATA_SOURCE.items():
            self.assertEqual(val, ds_db_obj.get(key),
                             "Key not found %s" % key)

    def test_data_source_delete(self):
        ctx = context.ctx()
        db_obj_ds = self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)
        _id = db_obj_ds['id']

        self.api.data_source_destroy(ctx, _id)

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.data_source_destroy(ctx, _id)

    def test_data_source_search(self):
        ctx = context.ctx()
        ctx.tenant_id = SAMPLE_DATA_SOURCE['tenant_id']
        self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)

        lst = self.api.data_source_get_all(ctx)
        self.assertEqual(1, len(lst))

        kwargs = {'name': SAMPLE_DATA_SOURCE['name'],
                  'tenant_id': SAMPLE_DATA_SOURCE['tenant_id']}
        lst = self.api.data_source_get_all(ctx, **kwargs)
        self.assertEqual(1, len(lst))

        # Valid field but no matching value
        kwargs = {'name': SAMPLE_DATA_SOURCE['name']+"foo"}
        lst = self.api.data_source_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Valid field with substrings
        kwargs = {'name': 'ngt',
                  'tenant_id': SAMPLE_DATA_SOURCE['tenant_id']}
        lst = self.api.data_source_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Invalid field
        self.assertRaises(sa_exc.InvalidRequestError,
                          self.api.data_source_get_all,
                          ctx, **{'badfield': 'somevalue'})

    @mock.patch('sahara.db.sqlalchemy.api.regex_filter')
    def test_data_source_search_regex(self, regex_filter):

        # do this so we can return the correct value
        def _regex_filter(query, cls, regex_cols, search_opts):
            return query, search_opts

        regex_filter.side_effect = _regex_filter

        ctx = context.ctx()
        self.api.data_source_get_all(ctx)
        self.assertEqual(0, regex_filter.call_count)

        self.api.data_source_get_all(ctx, regex_search=True, name="fox")
        self.assertEqual(1, regex_filter.call_count)
        args, kwargs = regex_filter.call_args
        self.assertIs(args[1], m.DataSource)
        self.assertEqual(args[2], ["name", "description", "url"])
        self.assertEqual(args[3], {"name": "fox"})

    def test_data_source_count_in(self):
        ctx = context.ctx()
        ctx.tenant_id = SAMPLE_DATA_SOURCE['tenant_id']
        src = copy.copy(SAMPLE_DATA_SOURCE)
        self.api.data_source_create(ctx, src)

        cnt = self.api.data_source_count(ctx, name='ngt_test')
        self.assertEqual(1, cnt)

        cnt = self.api.data_source_count(ctx, name=('ngt_test',
                                                    'test2', 'test3'))
        self.assertEqual(1, cnt)

        cnt = self.api.data_source_count(ctx, name=('test1',
                                                    'test2', 'test3'))
        self.assertEqual(0, cnt)

        lst = self.api.data_source_get_all(ctx, name='ngt_test')
        myid = lst[0]['id']
        cnt = self.api.data_source_count(ctx,
                                         name=('ngt_test', 'test2', 'test3'),
                                         id=myid)
        self.assertEqual(1, cnt)

        cnt = self.api.data_source_count(ctx,
                                         name=('ngt_test', 'test2', 'test3'),
                                         id=(myid, '2'))
        self.assertEqual(1, cnt)

    def test_data_source_count_like(self):
        ctx = context.ctx()
        ctx.tenant_id = SAMPLE_DATA_SOURCE['tenant_id']
        src = copy.copy(SAMPLE_DATA_SOURCE)
        self.api.data_source_create(ctx, src)

        cnt = self.api.data_source_count(ctx, name='ngt_test')
        self.assertEqual(1, cnt)

        cnt = self.api.data_source_count(ctx, name='ngt%')
        self.assertEqual(1, cnt)

        cnt = self.api.data_source_count(ctx,
                                         name=('ngt_test',),
                                         url='localhost%')
        self.assertEqual(1, cnt)

        cnt = self.api.data_source_count(ctx,
                                         name=('ngt_test',),
                                         url='localhost')
        self.assertEqual(0, cnt)

    def test_data_source_update(self):
        ctx = context.ctx()
        orig = self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)
        update_json = {"name": "updatedName",
                       "url": "swift://updatedFakeUrl"}
        updated = self.api.data_source_update(ctx, orig["id"], update_json)
        self.assertEqual("updatedName", updated["name"])
        self.assertEqual("swift://updatedFakeUrl", updated["url"])

    def test_ds_update_delete_when_protected(self):
        ctx = context.ctx()
        sample = copy.deepcopy(SAMPLE_DATA_SOURCE)
        sample['is_protected'] = True
        ds = self.api.data_source_create(ctx, sample)
        ds_id = ds["id"]

        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                self.api.data_source_update(ctx, ds_id, {"name": "ds"})
            except ex.UpdateFailedException as e:
                self.assert_protected_resource_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                self.api.data_source_destroy(ctx, ds_id)
            except ex.DeletionFailed as e:
                self.assert_protected_resource_exception(e)
                raise e

        self.api.data_source_update(ctx, ds_id,
                                    {"name": "ds", "is_protected": False})

    def test_public_ds_update_delete_from_another_tenant(self):
        ctx = context.ctx()
        sample = copy.deepcopy(SAMPLE_DATA_SOURCE)
        sample['is_public'] = True
        ds = self.api.data_source_create(ctx, sample)
        ds_id = ds["id"]
        ctx.tenant_id = 'tenant_2'

        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                self.api.data_source_update(ctx, ds_id, {"name": "ds"})
            except ex.UpdateFailedException as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                self.api.data_source_destroy(ctx, ds_id)
            except ex.DeletionFailed as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e


class JobExecutionTest(test_base.ConductorManagerTestCase):
    def setUp(self):
        super(JobExecutionTest, self).setUp()
        castellan.validate_config()

    def test_crud_operation_create_list_delete_update(self):
        ctx = context.ctx()
        job = self.api.job_create(ctx, SAMPLE_JOB)
        ds_input = self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT = copy.copy(SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT['name'] = 'output'
        ds_output = self.api.data_source_create(ctx, SAMPLE_DATA_OUTPUT)

        SAMPLE_JOB_EXECUTION['job_id'] = job['id']
        SAMPLE_JOB_EXECUTION['input_id'] = ds_input['id']
        SAMPLE_JOB_EXECUTION['output_id'] = ds_output['id']

        self.api.job_execution_create(ctx, SAMPLE_JOB_EXECUTION)

        lst = self.api.job_execution_get_all(ctx)
        self.assertEqual(1, len(lst))

        count = self.api.job_execution_count(ctx)
        self.assertEqual(1, count)

        job_ex_id = lst[0]['id']

        self.assertIsNone(lst[0]['info'])
        new_info = {"status": edp.JOB_STATUS_PENDING}
        self.api.job_execution_update(ctx, job_ex_id, {'info': new_info})
        updated_job = self.api.job_execution_get(ctx, job_ex_id)
        self.assertEqual(new_info, updated_job['info'])
        self.assertEqual(SAMPLE_JOB_EXECUTION['start_time'],
                         updated_job['start_time'])

        self.api.job_execution_destroy(ctx, job_ex_id)

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.job_execution_update(ctx, job_ex_id, {'info': new_info})

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.job_execution_destroy(ctx, job_ex_id)

        lst = self.api.job_execution_get_all(ctx)
        self.assertEqual(0, len(lst))

    def test_crud_operation_on_configured_jobs(self):
        ctx = context.ctx()
        job = self.api.job_create(ctx, SAMPLE_JOB)
        ds_input = self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT = copy.copy(SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT['name'] = 'output'
        ds_output = self.api.data_source_create(ctx, SAMPLE_DATA_OUTPUT)

        SAMPLE_CONF_JOB_EXECUTION['job_id'] = job['id']
        SAMPLE_CONF_JOB_EXECUTION['input_id'] = ds_input['id']
        SAMPLE_CONF_JOB_EXECUTION['output_id'] = ds_output['id']

        self.api.job_execution_create(ctx, SAMPLE_CONF_JOB_EXECUTION)

        lst = self.api.job_execution_get_all(ctx)
        self.assertEqual(1, len(lst))

        job_ex = lst[0]
        configs = {
            'conf2': 'value_je',
            'conf3': 'value_je'
        }
        self.assertEqual(configs, job_ex['job_configs'])

    def test_null_data_sources(self):
        ctx = context.ctx()
        job = self.api.job_create(ctx, SAMPLE_JOB)

        SAMPLE_CONF_JOB_EXECUTION['job_id'] = job['id']
        SAMPLE_CONF_JOB_EXECUTION['input_id'] = None
        SAMPLE_CONF_JOB_EXECUTION['output_id'] = None

        id = self.api.job_execution_create(ctx,
                                           SAMPLE_CONF_JOB_EXECUTION)['id']
        job_exec = self.api.job_execution_get(ctx, id)

        self.assertIsNone(job_exec['input_id'])
        self.assertIsNone(job_exec['output_id'])

    def test_deletion_constraints_on_data_and_jobs(self):
        ctx = context.ctx()
        job = self.api.job_create(ctx, SAMPLE_JOB)
        ds_input = self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT = copy.copy(SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT['name'] = 'output'
        ds_output = self.api.data_source_create(ctx, SAMPLE_DATA_OUTPUT)

        SAMPLE_CONF_JOB_EXECUTION['job_id'] = job['id']
        SAMPLE_CONF_JOB_EXECUTION['input_id'] = ds_input['id']
        SAMPLE_CONF_JOB_EXECUTION['output_id'] = ds_output['id']

        self.api.job_execution_create(ctx, SAMPLE_CONF_JOB_EXECUTION)

        with testtools.ExpectedException(ex.DeletionFailed):
            self.api.data_source_destroy(ctx, ds_input['id'])
        with testtools.ExpectedException(ex.DeletionFailed):
            self.api.data_source_destroy(ctx, ds_output['id'])
        with testtools.ExpectedException(ex.DeletionFailed):
            self.api.job_destroy(ctx, job['id'])

    def test_job_execution_search(self):
        ctx = context.ctx()
        jvals = copy.copy(SAMPLE_JOB)
        jvals["name"] = "frederica"
        job = self.api.job_create(ctx, jvals)

        ds_input = self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT = copy.copy(SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT['name'] = 'output'
        ds_output = self.api.data_source_create(ctx, SAMPLE_DATA_OUTPUT)

        job_exec = copy.copy(SAMPLE_JOB_EXECUTION)

        job_exec['job_id'] = job['id']
        job_exec['input_id'] = ds_input['id']
        job_exec['output_id'] = ds_output['id']

        ctx.tenant_id = job_exec['tenant_id']
        self.api.job_execution_create(ctx, job_exec)

        lst = self.api.job_execution_get_all(ctx)
        self.assertEqual(1, len(lst))

        kwargs = {'tenant_id': job_exec['tenant_id']}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(1, len(lst))

        # Valid field but no matching value
        kwargs = {'job_id': job_exec['job_id']+"foo"}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Valid field with substrings
        kwargs = {'job.name': "red"}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Invalid field
        self.assertRaises(sa_exc.InvalidRequestError,
                          self.api.job_execution_get_all,
                          ctx, **{'badfield': 'somevalue'})

    @mock.patch('sahara.db.sqlalchemy.api.regex_filter')
    def test_job_execution_search_regex(self, regex_filter):

        # do this so we can return the correct value
        def _regex_filter(query, cls, regex_cols, search_opts):
            return query, search_opts

        regex_filter.side_effect = _regex_filter

        ctx = context.ctx()
        self.api.job_execution_get_all(ctx)
        self.assertEqual(0, regex_filter.call_count)

        self.api.job_execution_get_all(ctx, regex_search=True,
                                       **{"job.name": "fox",
                                          "cluster.name": "jack",
                                          "id": "124"})

        self.assertEqual(3, regex_filter.call_count)

        # First call, after externals were removed
        args, kwargs = regex_filter.call_args_list[0]
        self.assertIs(args[1], m.JobExecution)
        self.assertEqual(args[2], ["job.name", "cluster.name"])
        self.assertEqual(args[3], {"id": "124"})

        # Second call, looking for cluster.name
        args, kwargs = regex_filter.call_args_list[1]
        self.assertIs(args[1], m.Cluster)
        self.assertEqual(args[2], ["name"])
        self.assertEqual(args[3], {"name": "jack"})

        # Third call, looking for job.name
        args, kwargs = regex_filter.call_args_list[2]
        self.assertIs(args[1], m.Job)
        self.assertEqual(args[2], ["name"])
        self.assertEqual(args[3], {"name": "fox"})

    def test_job_execution_advanced_search(self):
        ctx = context.ctx()
        job = self.api.job_create(ctx, SAMPLE_JOB)
        ds_input = self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT = copy.copy(SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT['name'] = 'output'
        ds_output = self.api.data_source_create(ctx, SAMPLE_DATA_OUTPUT)

        # Create a cluster
        cl1 = self.api.cluster_create(ctx, test_clusters.SAMPLE_CLUSTER)

        # Create a second cluster with a different name
        cl2_vals = copy.copy(test_clusters.SAMPLE_CLUSTER)
        cl2_vals['name'] = 'test_cluster2'
        cl2 = self.api.cluster_create(ctx, cl2_vals)

        my_sample_job_exec = copy.copy(SAMPLE_JOB_EXECUTION)

        my_sample_job_exec['job_id'] = job['id']
        my_sample_job_exec['input_id'] = ds_input['id']
        my_sample_job_exec['output_id'] = ds_output['id']
        my_sample_job_exec['cluster_id'] = cl1['id']

        # Run job on cluster 1
        self.api.job_execution_create(ctx, my_sample_job_exec)

        # Run the same job on cluster 2 and set status
        my_sample_job_exec['cluster_id'] = cl2['id']
        my_sample_job_exec['info'] = {'status': 'KiLLeD'}
        self.api.job_execution_create(ctx, my_sample_job_exec)

        # Search only with job execution fields (finds both)
        lst = self.api.job_execution_get_all(ctx, **{'return_code': 1})
        self.assertEqual(2, len(lst))

        # Search on cluster name
        kwargs = {'cluster.name': cl1['name'],
                  'return_code': 1}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(1, len(lst))

        # Search on cluster name and job name
        kwargs = {'cluster.name': cl1['name'],
                  'job.name': SAMPLE_JOB['name'],
                  'return_code': 1}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(1, len(lst))

        # Search on cluster name, job name, and status
        kwargs = {'cluster.name': cl2['name'],
                  'job.name': SAMPLE_JOB['name'],
                  'status': 'killed',
                  'return_code': 1}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(1, len(lst))

        # Search on job name (finds both)
        kwargs = {'job.name': SAMPLE_JOB['name'],
                  'return_code': 1}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(2, len(lst))

        # invalid cluster name value
        kwargs = {'cluster.name': cl1['name']+'foo',
                  'job.name': SAMPLE_JOB['name']}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # invalid job name value
        kwargs = {'cluster.name': cl1['name'],
                  'job.name': SAMPLE_JOB['name']+'foo'}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # invalid status value
        kwargs = {'cluster.name': cl1['name'],
                  'status': 'PENDING'}
        lst = self.api.job_execution_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))


class JobTest(test_base.ConductorManagerTestCase):
    def __init__(self, *args, **kwargs):
        super(JobTest, self).__init__(
            checks=[
                lambda: SAMPLE_JOB
            ], *args, **kwargs)

    def test_crud_operation_create_list_delete_update(self):
        ctx = context.ctx()

        self.api.job_create(ctx, SAMPLE_JOB)

        lst = self.api.job_get_all(ctx)
        self.assertEqual(1, len(lst))

        jo_id = lst[0]['id']

        update_jo = self.api.job_update(ctx, jo_id,
                                        {'description': 'update'})
        self.assertEqual('update', update_jo['description'])

        self.api.job_destroy(ctx, jo_id)

        lst = self.api.job_get_all(ctx)
        self.assertEqual(0, len(lst))

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.job_destroy(ctx, jo_id)

    def test_job_fields(self):
        ctx = context.ctx()
        ctx.tenant_id = SAMPLE_JOB['tenant_id']
        job_id = self.api.job_create(ctx, SAMPLE_JOB)['id']

        job = self.api.job_get(ctx, job_id)
        self.assertIsInstance(job, dict)

        for key, val in SAMPLE_JOB.items():
            self.assertEqual(val, job.get(key),
                             "Key not found %s" % key)

    def test_job_search(self):
        ctx = context.ctx()
        job = copy.copy(SAMPLE_JOB)
        job["name"] = "frederica"
        job["description"] = "thebestjob"
        ctx.tenant_id = job['tenant_id']
        self.api.job_create(ctx, job)

        lst = self.api.job_get_all(ctx)
        self.assertEqual(1, len(lst))

        kwargs = {'name': job['name'],
                  'tenant_id': job['tenant_id']}
        lst = self.api.job_get_all(ctx, **kwargs)
        self.assertEqual(1, len(lst))

        # Valid field but no matching value
        lst = self.api.job_get_all(ctx, **{'name': job['name']+"foo"})
        self.assertEqual(0, len(lst))

        # Valid field with substrings
        kwargs = {'name': "red",
                  'description': "best"}
        lst = self.api.job_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Invalid field
        self.assertRaises(sa_exc.InvalidRequestError,
                          self.api.job_get_all,
                          ctx, **{'badfield': 'somevalue'})

    @mock.patch('sahara.db.sqlalchemy.api.regex_filter')
    def test_job_search_regex(self, regex_filter):

        # do this so we can return the correct value
        def _regex_filter(query, cls, regex_cols, search_opts):
            return query, search_opts

        regex_filter.side_effect = _regex_filter

        ctx = context.ctx()
        self.api.job_get_all(ctx)
        self.assertEqual(0, regex_filter.call_count)

        self.api.job_get_all(ctx, regex_search=True, name="fox")
        self.assertEqual(1, regex_filter.call_count)
        args, kwargs = regex_filter.call_args
        self.assertIs(args[1], m.Job)
        self.assertEqual(args[2], ["name", "description"])
        self.assertEqual(args[3], {"name": "fox"})

    def test_job_update_delete_when_protected(self):
        ctx = context.ctx()
        sample = copy.deepcopy(SAMPLE_JOB)
        sample['is_protected'] = True
        job = self.api.job_create(ctx, sample)
        job_id = job["id"]

        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                self.api.job_update(ctx, job_id, {"name": "job"})
            except ex.UpdateFailedException as e:
                self.assert_protected_resource_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                self.api.job_destroy(ctx, job_id)
            except ex.DeletionFailed as e:
                self.assert_protected_resource_exception(e)
                raise e

        self.api.job_update(ctx, job_id, {"name": "job",
                                          "is_protected": False})

    def test_public_job_update_delete_from_another_tenant(self):
        ctx = context.ctx()
        sample = copy.deepcopy(SAMPLE_JOB)
        sample['is_public'] = True
        job = self.api.job_create(ctx, sample)
        job_id = job["id"]
        ctx.tenant_id = 'tenant_2'

        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                self.api.job_update(ctx, job_id, {"name": "job"})
            except ex.UpdateFailedException as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                self.api.job_destroy(ctx, job_id)
            except ex.DeletionFailed as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e


class JobBinaryInternalTest(test_base.ConductorManagerTestCase):
    def __init__(self, *args, **kwargs):
        super(JobBinaryInternalTest, self).__init__(
            checks=[
                lambda: SAMPLE_JOB_BINARY_INTERNAL
            ], *args, **kwargs)

    def test_crud_operation_create_list_delete_update(self):
        ctx = context.ctx()

        self.api.job_binary_internal_create(ctx, SAMPLE_JOB_BINARY_INTERNAL)

        lst = self.api.job_binary_internal_get_all(ctx)
        self.assertEqual(1, len(lst))

        job_bin_int_id = lst[0]['id']

        update_jbi = self.api.job_binary_internal_update(
            ctx, job_bin_int_id, {'name': 'newname'})

        self.assertEqual('newname', update_jbi['name'])

        self.api.job_binary_internal_destroy(ctx, job_bin_int_id)

        lst = self.api.job_binary_internal_get_all(ctx)
        self.assertEqual(0, len(lst))

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.job_binary_internal_destroy(ctx, job_bin_int_id)

    def test_duplicate_job_binary_internal_create(self):
        ctx = context.ctx()
        self.api.job_binary_internal_create(ctx, SAMPLE_JOB_BINARY_INTERNAL)
        with testtools.ExpectedException(ex.DBDuplicateEntry):
            self.api.job_binary_internal_create(ctx,
                                                SAMPLE_JOB_BINARY_INTERNAL)

    def test_job_binary_internal_get_raw(self):
        ctx = context.ctx()

        id = self.api.job_binary_internal_create(ctx,
                                                 SAMPLE_JOB_BINARY_INTERNAL
                                                 )['id']
        data = self.api.job_binary_internal_get_raw_data(ctx, id)
        self.assertEqual(SAMPLE_JOB_BINARY_INTERNAL["data"], data)

        self.api.job_binary_internal_destroy(ctx, id)

        data = self.api.job_binary_internal_get_raw_data(ctx, id)
        self.assertIsNone(data)

    def test_job_binary_internal_fields(self):
        ctx = context.ctx()
        ctx.tenant_id = SAMPLE_JOB_BINARY_INTERNAL['tenant_id']
        id = self.api.job_binary_internal_create(
            ctx, SAMPLE_JOB_BINARY_INTERNAL)['id']

        internal = self.api.job_binary_internal_get(ctx, id)
        self.assertIsInstance(internal, dict)
        with testtools.ExpectedException(KeyError):
            internal["data"]

        internal["data"] = self.api.job_binary_internal_get_raw_data(ctx, id)
        for key, val in SAMPLE_JOB_BINARY_INTERNAL.items():
            if key == "datasize":
                self.assertEqual(len(BINARY_DATA), internal["datasize"])
            else:
                self.assertEqual(val, internal.get(key),
                                 "Key not found %s" % key)

    def test_job_binary_internal_search(self):
        ctx = context.ctx()
        jbi = copy.copy(SAMPLE_JOB_BINARY_INTERNAL)
        jbi["name"] = "frederica"
        ctx.tenant_id = jbi['tenant_id']
        self.api.job_binary_internal_create(ctx, jbi)

        lst = self.api.job_binary_internal_get_all(ctx)
        self.assertEqual(1, len(lst))

        kwargs = {'name': jbi['name'],
                  'tenant_id': jbi['tenant_id']}
        lst = self.api.job_binary_internal_get_all(ctx, **kwargs)
        self.assertEqual(1, len(lst))

        # Valid field but no matching value
        kwargs = {'name': jbi['name']+"foo"}
        lst = self.api.job_binary_internal_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Valid field with substrings
        kwargs = {'name': "red"}
        lst = self.api.job_binary_internal_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Invalid field
        self.assertRaises(sa_exc.InvalidRequestError,
                          self.api.job_binary_internal_get_all,
                          ctx, **{'badfield': 'junk'})

    @mock.patch('sahara.db.sqlalchemy.api.regex_filter')
    def test_job_binary_internal_search_regex(self, regex_filter):

        # do this so we can return the correct value
        def _regex_filter(query, cls, regex_cols, search_opts):
            return query, search_opts

        regex_filter.side_effect = _regex_filter

        ctx = context.ctx()
        self.api.job_binary_internal_get_all(ctx)
        self.assertEqual(0, regex_filter.call_count)

        self.api.job_binary_internal_get_all(ctx,
                                             regex_search=True, name="fox")
        self.assertEqual(1, regex_filter.call_count)
        args, kwargs = regex_filter.call_args
        self.assertIs(args[1], m.JobBinaryInternal)
        self.assertEqual(args[2], ["name"])
        self.assertEqual(args[3], {"name": "fox"})

    def test_jbi_update_delete_when_protected(self):
        ctx = context.ctx()
        sample = copy.deepcopy(SAMPLE_JOB_BINARY_INTERNAL)
        sample['is_protected'] = True
        jbi = self.api.job_binary_internal_create(ctx, sample)
        jbi_id = jbi["id"]

        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                self.api.job_binary_internal_update(ctx, jbi_id,
                                                    {"name": "jbi"})
            except ex.UpdateFailedException as e:
                self.assert_protected_resource_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                self.api.job_binary_internal_destroy(ctx, jbi_id)
            except ex.DeletionFailed as e:
                self.assert_protected_resource_exception(e)
                raise e

        self.api.job_binary_internal_update(ctx, jbi_id,
                                            {"name": "jbi",
                                             "is_protected": False})

    def test_public_jbi_update_delete_from_another_tenant(self):
        ctx = context.ctx()
        sample = copy.deepcopy(SAMPLE_JOB_BINARY_INTERNAL)
        sample['is_public'] = True
        jbi = self.api.job_binary_internal_create(ctx, sample)
        jbi_id = jbi["id"]
        ctx.tenant_id = 'tenant_2'

        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                self.api.job_binary_internal_update(ctx, jbi_id,
                                                    {"name": "jbi"})
            except ex.UpdateFailedException as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                self.api.job_binary_internal_destroy(ctx, jbi_id)
            except ex.DeletionFailed as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e


class JobBinaryTest(test_base.ConductorManagerTestCase):
    def __init__(self, *args, **kwargs):
        super(JobBinaryTest, self).__init__(
            checks=[
                lambda: SAMPLE_JOB_BINARY
            ], *args, **kwargs)

    def setUp(self):
        super(JobBinaryTest, self).setUp()
        castellan.validate_config()

    def test_crud_operation_create_list_delete(self):
        ctx = context.ctx()

        self.api.job_binary_create(ctx, SAMPLE_JOB_BINARY)

        lst = self.api.job_binary_get_all(ctx)
        self.assertEqual(1, len(lst))

        job_binary_id = lst[0]['id']
        self.api.job_binary_destroy(ctx, job_binary_id)

        lst = self.api.job_binary_get_all(ctx)
        self.assertEqual(0, len(lst))

        with testtools.ExpectedException(ex.NotFoundException):
            self.api.job_binary_destroy(ctx, job_binary_id)

    def test_job_binary_fields(self):
        ctx = context.ctx()
        ctx.tenant_id = SAMPLE_JOB_BINARY['tenant_id']
        job_binary_id = self.api.job_binary_create(ctx,
                                                   SAMPLE_JOB_BINARY)['id']

        job_binary = self.api.job_binary_get(ctx, job_binary_id)
        self.assertIsInstance(job_binary, dict)

        for key, val in SAMPLE_JOB_BINARY.items():
            self.assertEqual(val, job_binary.get(key),
                             "Key not found %s" % key)

    def _test_job_binary_referenced(self, reference):
        ctx = context.ctx()
        job_binary_id = self.api.job_binary_create(ctx,
                                                   SAMPLE_JOB_BINARY)['id']

        job_values = copy.copy(SAMPLE_JOB)
        job_values[reference] = [job_binary_id]
        job_id = self.api.job_create(ctx, job_values)['id']

        # Delete while referenced, fails
        with testtools.ExpectedException(ex.DeletionFailed):
            self.api.job_binary_destroy(ctx, job_binary_id)

        # Delete while not referenced
        self.api.job_destroy(ctx, job_id)
        self.api.job_binary_destroy(ctx, job_binary_id)
        lst = self.api.job_binary_get_all(ctx)
        self.assertEqual(0, len(lst))

    def test_job_binary_referenced_mains(self):
        self._test_job_binary_referenced("mains")

    def test_job_binary_referenced_libs(self):
        self._test_job_binary_referenced("libs")

    def test_duplicate_job_binary_create(self):
        ctx = context.ctx()
        self.api.job_binary_create(ctx, SAMPLE_JOB_BINARY)
        with testtools.ExpectedException(ex.DBDuplicateEntry):
            self.api.job_binary_create(ctx,
                                       SAMPLE_JOB_BINARY)

    def test_job_binary_search(self):
        ctx = context.ctx()
        jb = copy.copy(SAMPLE_JOB_BINARY)
        jb["name"] = "frederica"
        jb["url"] = "http://thebestbinary"
        ctx.tenant_id = jb['tenant_id']
        self.api.job_binary_create(ctx, jb)

        lst = self.api.job_binary_get_all(ctx)
        self.assertEqual(1, len(lst))

        kwargs = {'name': jb['name'],
                  'tenant_id': jb['tenant_id']}
        lst = self.api.job_binary_get_all(ctx, **kwargs)
        self.assertEqual(1, len(lst))

        # Valid field but no matching value
        kwargs = {'name': jb['name']+"foo"}
        lst = self.api.job_binary_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Valid field with substrings
        kwargs = {'name': "red",
                  'url': "best"}
        lst = self.api.job_binary_get_all(ctx, **kwargs)
        self.assertEqual(0, len(lst))

        # Invalid field
        self.assertRaises(sa_exc.InvalidRequestError,
                          self.api.job_binary_get_all,
                          ctx, **{'badfield': 'somevalue'})

    @mock.patch('sahara.db.sqlalchemy.api.regex_filter')
    def test_job_binary_search_regex(self, regex_filter):

        # do this so we can return the correct value
        def _regex_filter(query, cls, regex_cols, search_opts):
            return query, search_opts

        regex_filter.side_effect = _regex_filter

        ctx = context.ctx()
        self.api.job_binary_get_all(ctx)
        self.assertEqual(0, regex_filter.call_count)

        self.api.job_binary_get_all(ctx, regex_search=True, name="fox")
        self.assertEqual(1, regex_filter.call_count)
        args, kwargs = regex_filter.call_args
        self.assertIs(args[1], m.JobBinary)
        self.assertEqual(args[2], ["name", "description", "url"])
        self.assertEqual(args[3], {"name": "fox"})

    def test_job_binary_update(self):
        ctx = context.ctx()

        original = self.api.job_binary_create(ctx, SAMPLE_JOB_BINARY_SWIFT)
        updated = self.api.job_binary_update(
            ctx, original["id"], SAMPLE_JOB_BINARY_SWIFT_UPDATE)
        # Make sure that the update did indeed succeed
        self.assertEqual(
            SAMPLE_JOB_BINARY_SWIFT_UPDATE["name"], updated["name"])
        self.assertEqual(SAMPLE_JOB_BINARY_SWIFT_UPDATE["url"], updated["url"])

        # Make sure we do NOT update a binary in use by a PENDING job
        self._create_job_execution_ref_job_binary(ctx, original["id"])
        with testtools.ExpectedException(ex.UpdateFailedException):
            self.api.job_binary_update(
                ctx, original["id"], SAMPLE_JOB_BINARY_SWIFT_UPDATE)

        original = self.api.job_binary_create(ctx, SAMPLE_JOB_BINARY)
        # Make sure that internal URL update fails
        with testtools.ExpectedException(ex.UpdateFailedException):
            self.api.job_binary_update(
                ctx, original["id"], SAMPLE_JOB_BINARY_UPDATE)

    def _create_job_execution_ref_job_binary(self, ctx, jb_id):
        JOB_REF_BINARY = copy.copy(SAMPLE_JOB)
        JOB_REF_BINARY["mains"] = [jb_id]
        job = self.api.job_create(ctx, JOB_REF_BINARY)
        ds_input = self.api.data_source_create(ctx, SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT = copy.copy(SAMPLE_DATA_SOURCE)
        SAMPLE_DATA_OUTPUT['name'] = 'output'
        ds_output = self.api.data_source_create(ctx, SAMPLE_DATA_OUTPUT)
        SAMPLE_JOB_EXECUTION['job_id'] = job['id']
        SAMPLE_JOB_EXECUTION['input_id'] = ds_input['id']
        SAMPLE_JOB_EXECUTION['output_id'] = ds_output['id']

        self.api.job_execution_create(ctx, SAMPLE_JOB_EXECUTION)
        lst = self.api.job_execution_get_all(ctx)
        job_ex_id = lst[0]["id"]
        new_info = {"status": edp.JOB_STATUS_PENDING}
        self.api.job_execution_update(ctx, job_ex_id, {"info": new_info})

    def test_jb_update_delete_when_protected(self):
        ctx = context.ctx()
        sample = copy.deepcopy(SAMPLE_JOB_BINARY)
        sample['is_protected'] = True
        jb = self.api.job_binary_create(ctx, sample)
        jb_id = jb["id"]

        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                self.api.job_binary_update(ctx, jb_id, {"name": "jb"})
            except ex.UpdateFailedException as e:
                self.assert_protected_resource_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                self.api.job_binary_destroy(ctx, jb_id)
            except ex.DeletionFailed as e:
                self.assert_protected_resource_exception(e)
                raise e

        self.api.job_binary_update(ctx, jb_id, {"name": "jb",
                                                "is_protected": False})

    def test_public_jb_update_delete_from_another_tenant(self):
        ctx = context.ctx()
        sample = copy.deepcopy(SAMPLE_JOB_BINARY)
        sample['is_public'] = True
        jb = self.api.job_binary_create(ctx, sample)
        jb_id = jb["id"]
        ctx.tenant_id = 'tenant_2'

        with testtools.ExpectedException(ex.UpdateFailedException):
            try:
                self.api.job_binary_update(ctx, jb_id, {"name": "jb"})
            except ex.UpdateFailedException as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e

        with testtools.ExpectedException(ex.DeletionFailed):
            try:
                self.api.job_binary_destroy(ctx, jb_id)
            except ex.DeletionFailed as e:
                self.assert_created_in_another_tenant_exception(e)
                raise e



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from sahara.swift import swift_helper as h
from sahara.tests.unit import base


GENERAL_PREFIX = "fs.swift."
SERVICE_PREFIX = "service.sahara."

GENERAL = ["impl", "connect.timeout", "socket.timeout",
           "connect.retry.count", "connect.throttle.delay",
           "blocksize", "partsize", "requestsize"]

SERVICE_SPECIFIC = ["auth.url", "tenant",
                    "username", "password", "http.port",
                    "https.port", "public", "location-aware",
                    "region", "apikey"]


class SwiftIntegrationTestCase(base.SaharaTestCase):
    @mock.patch('sahara.utils.openstack.base.retrieve_auth_url')
    def test_get_swift_configs(self, url_for_mock):
        url_for_mock.return_value = 'http://localhost:5000/v2.0'
        self.setup_context(tenant_name='test_tenant')
        self.override_config("os_region_name", 'regionOne')

        result = h.get_swift_configs()
        self.assertEqual(8, len(result))
        self.assertIn({'name': "fs.swift.service.sahara.tenant",
                       'value': 'test_tenant', 'description': ''}, result)
        self.assertIn({'name': "fs.swift.service.sahara.http.port",
                       'value': '8080', 'description': ''}, result)
        self.assertIn({'name': "fs.swift.service.sahara.region",
                       'value': 'regionOne', 'description': ''}, result)



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools

import sahara.exceptions as ex
from sahara.service.edp.data_sources.maprfs.implementation import MapRFSType
from sahara.tests.unit import base


class TestMapRFSTypeValidation(base.SaharaTestCase):
    def setUp(self):
        super(TestMapRFSTypeValidation, self).setUp()
        self.maprfs_type = MapRFSType()

    def test_maprfs_type_validation_wrong_schema(self):
        data = {
            "name": "test_data_data_source",
            "url": "maprf://test_cluster/",
            "type": "maprfs",
            "description": "incorrect url schema"
        }
        with testtools.ExpectedException(ex.InvalidDataException):
            self.maprfs_type.validate(data)

    def test_maprfs_type_validation_correct_url(self):
        data = {
            "name": "test_data_data_source",
            "url": "maprfs:///test_cluster/",
            "type": "maprfs",
            "description": "correct url schema"
        }
        self.maprfs_type.validate(data)

    def test_maprfs_type_validation_local_rel_url(self):
        data = {
            "name": "test_data_data_source",
            "url": "mydata/input",
            "type": "maprfs",
            "description": ("correct url schema for"
                            " relative path on local maprfs")
        }
        self.maprfs_type.validate(data)

    def test_maprfs_type_validation_local_abs_url(self):
        data = {
            "name": "test_data_data_source",
            "url": "/tmp/output",
            "type": "maprfs",
            "description": ("correct url schema for"
                            " absolute path on local maprfs")
        }
        self.maprfs_type.validate(data)



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_utils import uuidutils

from sahara.conductor import resource as r


def create_cluster(name, tenant, plugin, version, node_groups, **kwargs):
    dct = {'id': uuidutils.generate_uuid(), 'name': name,
           'tenant_id': tenant, 'plugin_name': plugin,
           'hadoop_version': version, 'node_groups': node_groups,
           'cluster_configs': {}, "sahara_info": {},
           'user_keypair_id': None, 'default_image_id': None,
           'is_protected': False}
    dct.update(kwargs)
    return r.ClusterResource(dct)


def make_ng_dict(name, flavor, processes, count, instances=None,
                 volumes_size=None, node_configs=None, resource=False,
                 **kwargs):
    node_configs = node_configs or {}
    instances = instances or []
    dct = {'id': uuidutils.generate_uuid(), 'name': name,
           'volumes_size': volumes_size, 'flavor_id': flavor,
           'node_processes': processes, 'count': count,
           'instances': instances, 'node_configs': node_configs,
           'security_groups': None, 'auto_security_group': False,
           'availability_zone': None, 'volumes_availability_zone': None,
           'open_ports': [], 'is_proxy_gateway': False,
           'volume_local_to_instance': False}
    dct.update(kwargs)
    if resource:
        return r.NodeGroupTemplateResource(dct)
    return dct


def make_inst_dict(inst_id, inst_name, management_ip='1.2.3.4'):
    return {'instance_id': inst_id, 'instance_name': inst_name,
            'management_ip': management_ip}



# Copyright (c) 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from sahara import conductor as c
from sahara import context
from sahara import exceptions as ex
from sahara.service import api
from sahara.service.edp import job_manager as manager
from sahara.utils import edp
from sahara.utils import proxy as p


conductor = c.API
LOG = logging.getLogger(__name__)


def execute_job(data):
    # Elements common to all job types
    job_template_id = data['job_template_id']
    cluster_id = data['cluster_id']
    configs = data.get('job_configs', {})
    interface = data.get('interface', {})

    # Not in Java job types but present for all others
    input_id = data.get('input_id', None)
    output_id = data.get('output_id', None)

    # Since we will use a unified class in the database, we pass
    # a superset for all job types
    # example configs['start'] = '2015-05-12T08:55Z' frequency = 5 mins
    # the job will starts from 2015-05-12T08:55Z, runs every 5 mins

    job_execution_info = data.get('job_execution_info', {})

    configs['job_execution_info'] = job_execution_info

    job_ex_dict = {'input_id': input_id, 'output_id': output_id,
                   'job_id': job_template_id, 'cluster_id': cluster_id,
                   'info': {'status': edp.JOB_STATUS_PENDING},
                   'job_configs': configs, 'extra': {},
                   'interface': interface}
    job_execution = conductor.job_execution_create(context.ctx(), job_ex_dict)
    context.set_current_job_execution_id(job_execution.id)

    # check to use proxy user
    if p.job_execution_requires_proxy_user(job_execution):
        try:
            p.create_proxy_user_for_job_execution(job_execution)
        except ex.SaharaException as e:
            LOG.error("Can't run job execution. "
                      "(Reasons: {reason})".format(reason=e))
            conductor.job_execution_destroy(context.ctx(), job_execution)
            raise e

    api.OPS.run_edp_job(job_execution.id)

    return job_execution


def job_execution_list(**kwargs):
    return conductor.job_execution_get_all(context.ctx(),
                                           regex_search=True, **kwargs)


def get_job_execution(id, refresh_status=False):
    if refresh_status:
        return manager.get_job_status(id)
    else:
        return conductor.job_execution_get(context.ctx(), id)


def update_job_execution(id, values):
    _update_status(values.pop("info", None), id)
    return conductor.job_execution_update(context.ctx(), id, values)


def _update_status(info, id):
    if info:
        status = info.get("status", None)
        if status == edp.JOB_ACTION_SUSPEND:
            api.OPS.job_execution_suspend(id)
        if status == edp.JOB_ACTION_CANCEL:
            api.OPS.cancel_job_execution(id)


def delete_job_execution(id):
    context.set_current_job_execution_id(id)
    api.OPS.delete_job_execution(id)



# Copyright (c) 2018 Massachusetts Open Cloud
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from oslo_middleware import base
from oslo_serialization import jsonutils
import webob
import webob.dec

from sahara.api import microversion as mv


class VersionResponseMiddlewareV1(base.Middleware):

    @webob.dec.wsgify
    def __call__(self, req):
        """Respond to a request for all Sahara API versions."""
        path = req.environ['PATH_INFO']
        if re.match(r"^/*$", path):
            response = webob.Response(request=req, status=300,
                                      content_type="application/json")
            response.body = jsonutils.dump_as_bytes(self._get_versions(req))
            return response
        else:
            return self.application

    def _get_versions(self, req):
        """Populate the version response with APIv1 stuff."""
        version_response = {
            "versions": [
                {"id": "v1.0",
                 "status": "SUPPORTED",
                 "links": self._get_links("1.0", req)
                 },
                {"id": "v1.1",
                 "status": "CURRENT",
                 "links": self._get_links("1.1", req)
                 }
            ]
        }
        return version_response

    @staticmethod
    def _get_links(version, req):
        href = "%s/v%s/" % (req.host_url, version)
        return [{"rel": "self", "href": href}]


class VersionResponseMiddlewareV2(VersionResponseMiddlewareV1):

    def _get_versions(self, req):
        """Populate the version response with APIv1 and APIv2 stuff."""
        version_response = (
            super(VersionResponseMiddlewareV2, self)._get_versions(req)
        )
        version_response["versions"][1]["status"] = "SUPPORTED"  # v1.1
        version_response["versions"].append(
            {"id": "v2",
             "status": "CURRENT",
             "links": self._get_links("2", req),
             "min_version": mv.MIN_API_VERSION,
             "max_version": mv.MAX_API_VERSION
             }
        )
        return version_response



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from sahara.service.validations.edp import job_interface as j_i
from sahara.utils import edp


JOB_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "minLength": 1,
            "maxLength": 50,
            "format": "valid_name"
        },
        "description": {
            "type": "string"
        },
        "type": {
            "type": "string",
            "enum": edp.JOB_TYPES_ALL,
        },
        "mains": {
            "type": "array",
            "uniqueItems": True,
            "items": {
                "type": "string",
                "minLength": 1,
            }
        },
        "libs": {
            "type": "array",
            "uniqueItems": True,
            "items": {
                "type": "string",
                "minLength": 1,
            }
        },
        "streaming": {
            "type": "boolean"
        },
        "interface": j_i.INTERFACE_ARGUMENT_SCHEMA,
        "is_public": {
            "type": ["boolean", "null"],
        },
        "is_protected": {
            "type": ["boolean", "null"],
        }
    },
    "additionalProperties": False,
    "required": [
        "name",
        "type",
    ]
}


JOB_UPDATE_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "minLength": 1,
            "maxLength": 50,
            "format": "valid_name"
        },
        "description": {
            "type": ["string", "null"]
        },
        "is_public": {
            "type": ["boolean", "null"],
        },
        "is_protected": {
            "type": ["boolean", "null"],
        }
    },
    "additionalProperties": False,
    "required": []
}



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import six

from sahara import conductor


def name_and_id(template):
    return "{name} ({id})".format(name=template["name"],
                                  id=template["id"])


def is_node_group(template):
    # Node group templates and cluster templates have
    # different required fields in validation and neither
    # allows additional fields.  So, the presence of
    # node_processes or flavor_id should suffice to
    # identify a node group template. Check for both
    # to be nice, in case someone made a typo.
    return 'node_processes' in template or 'flavor_id' in template


def substitute_ng_ids(cl, ng_dict):
    '''Substitute node group template ids for node group template names

    If the cluster template contains node group elements with
    node_group_template_id fields that reference node group templates
    by name, substitute the node group template id for the name.
    The name reference is expected to be a string containing a format
    specifier of the form "{name}", for example "{master}"

    :param cl: a cluster template
    :param ng_dict: a dictionary of node group template ids keyed by
                    node group template names
    '''
    for ng in cl["node_groups"]:
        if "node_group_template_id" in ng:
            val = ng["node_group_template_id"].format(**ng_dict)
            ng["node_group_template_id"] = val


def check_basic_fields(template):
    return "plugin_name" in template and (
        "hadoop_version" in template and (
            "name" in template))


def check_plugin_version(template, plugin_versions):
    '''Check that the template  matches the plugin versions list

    Tests whether or not the plugin version indicated by the template
    matches one of the versions specified in plugin_versions

    :param template: A node group or cluster template
    :param plugin_versions: A list of plugin version strings. These
                            values may be regular version strings or may be
                            the name of  the plugin followed by a
                            "." followed by a version string.
    :returns: True if the plugin version specified in the template
              matches a version in plugin_versions or plugin_versions
              is an empty list. Otherwise False
    '''
    def dotted_name(template):
        return template['plugin_name'] + "." + template['hadoop_version']

    version_matches = plugin_versions is None or (
        template['hadoop_version'] in plugin_versions) or (
            dotted_name(template) in plugin_versions)

    return version_matches


def check_plugin_name_and_version(template, plugin_names, plugin_versions):
    '''Check that the template is for one of the specified plugins

    Tests whether or not the plugin name and version indicated by the template
    matches one of the names and one of the versions specified in
    plugin_names and plugin_versions

    :param template: A node group or cluster template
    :param plugin_names: A list of plugin names
    :param plugin_versions: A list of plugin version strings. These
                            values may be regular version strings or may be
                            the name of  the plugin followed by a
                            "." followed by a version string.
    :returns: True if the plugin name specified in the template matches
              a name in plugin_names or plugin_names is an empty list, and if
              the plugin version specified in the template matches a version
              in plugin_versions or plugin_versions is an empty list.
              Otherwise False
    '''
    name_and_version_matches = (plugin_names is None or (
        template['plugin_name'] in plugin_names)) and (
            check_plugin_version(template, plugin_versions))

    return name_and_version_matches


# TODO(tmckay): refactor the service validation code so
# that the node group template usage checks there can be reused
# without incurring unnecessary dependencies
def check_node_group_template_usage(node_group_template_id,
                                    cluster_list, cluster_template_list=None):
    cluster_template_list = cluster_template_list or []
    cluster_users = []
    template_users = []

    for cluster in cluster_list:
        if (node_group_template_id in
            [node_group.node_group_template_id
             for node_group in cluster.node_groups]):
            cluster_users += [cluster.name]

    for cluster_template in cluster_template_list:
        if (node_group_template_id in
            [node_group.node_group_template_id
             for node_group in cluster_template.node_groups]):
            template_users += [cluster_template.name]

    return cluster_users, template_users


# TODO(tmckay): refactor the service validation code so
# that the cluster template usage checks there can be reused
# without incurring unnecessary dependencies
def check_cluster_template_usage(cluster_template_id, cluster_list):
    cluster_users = []
    for cluster in cluster_list:
        if cluster_template_id == cluster.cluster_template_id:
            cluster_users.append(cluster.name)

    return cluster_users


def find_node_group_template_by_name(ctx, name):
    t = conductor.API.node_group_template_get_all(ctx,
                                                  name=name,
                                                  is_default=True)
    if t:
        return t[0]
    return None


def find_cluster_template_by_name(ctx, name):
    t = conductor.API.cluster_template_get_all(ctx,
                                               name=name,
                                               is_default=True)
    if t:
        return t[0]
    return None


def value_diff(current, new_values):
    '''Return the entries in current that would be overwritten by new_values

    Returns the set of entries in current that would be overwritten
    if current.update(new_values) was called.

    :param current: A dictionary whose key values are a superset
                    of the key values in new_values
    :param new_values: A dictionary
    '''
    # Current is an existing template from the db and
    # template is a set of values that has been validated
    # against the JSON schema for the template.
    # Copy items from current if they are present in template.

    # In the case of "node_groups" the conductor does magic
    # to set up template relations and insures that appropriate
    # fields are cleaned (like "updated_at" and "id") so we
    # trust the conductor in that case.

    diff_values = {}
    for k, v in six.iteritems(new_values):
        if k in current and current[k] != v:
            diff_values[k] = copy.deepcopy(current[k])
    return diff_values



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import testtools

from sahara import conductor
from sahara import context
from sahara import exceptions
from sahara.tests.unit import base
from sahara.utils import general as gu


SAMPLE_CLUSTER = {
    'plugin_name': 'test_plugin',
    'hadoop_version': 'test_version',
    'tenant_id': 'tenant_1',
    'name': 'test_cluster',
    'user_keypair_id': 'my_keypair',
    'node_groups': [
        {
            'name': 'ng_1',
            'flavor_id': '42',
            'node_processes': ['p1', 'p2'],
            'count': 1
        },
        {
            'name': 'ng_2',
            'flavor_id': '42',
            'node_processes': ['p3', 'p4'],
            'count': 3
        }
    ],
    'cluster_configs': {
        'service_1': {
            'config_2': 'value_2'
        },
        'service_2': {
            'config_1': 'value_1'
        }
    },
}

SAMPLE_NODE_GROUP = {
    'name': 'ng_3',
    'flavor_id': '42',
    'node_processes': ['p5', 'p6'],
    'count': 5
}

SAMPLE_INSTANCE = {
    'instance_name': 'test-name',
    'instance_id': '123456',
    'management_ip': '1.2.3.1'
}

SAMPLE_JOB = {
    "tenant_id": "test_tenant",
    "name": "job_test",
    "description": "test_desc",
    "type": "pig"
}

SAMPLE_JOB_BINARY = {
    "tenant_id": "test_tenant",
    "name": "job_binary_test",
    "description": "test_dec",
    "url": "internal-db://test_binary",
}


class TestConductorApi(base.SaharaWithDbTestCase):
    def setUp(self):
        super(TestConductorApi, self).setUp()
        self.api = conductor.API

    def _make_sample(self):
        ctx = context.ctx()
        cluster = self.api.cluster_create(ctx, SAMPLE_CLUSTER)
        return ctx, cluster

    def test_update_by_id(self):
        ctx, cluster = self._make_sample()

        self.api.cluster_update(ctx, cluster.id, {'name': 'changed'})

        updated_cluster = self.api.cluster_get(ctx, cluster.id)
        self.assertEqual('changed', updated_cluster['name'])

        self.api.cluster_destroy(ctx, updated_cluster.id)
        cluster_list = self.api.cluster_get_all(ctx)
        self.assertEqual(0, len(cluster_list))

    def test_add_node_group_to_cluster_id(self):
        ctx, cluster = self._make_sample()
        ng_id = self.api.node_group_add(ctx, cluster.id, SAMPLE_NODE_GROUP)
        self.assertTrue(ng_id)

    def test_update_node_group_by_id(self):
        ctx, cluster = self._make_sample()
        ng_id = cluster.node_groups[0].id
        self.api.node_group_update(ctx, ng_id, {'name': 'changed_ng'})
        cluster = self.api.cluster_get(ctx, cluster.id)

        ng = gu.get_by_id(cluster.node_groups, ng_id)
        self.assertEqual('changed_ng', ng.name)

    def test_remove_node_group(self):
        ctx, cluster = self._make_sample()
        ng = cluster.node_groups[0]
        self.api.node_group_remove(ctx, ng)

        cluster = self.api.cluster_get(ctx, cluster.id)
        self.assertNotIn(ng, cluster.node_groups)

    def test_add_instance_to_node_group_id(self):
        ctx, cluster = self._make_sample()
        inst_id = self.api.instance_add(ctx, cluster.node_groups[0].id,
                                        SAMPLE_INSTANCE)
        self.assertTrue(inst_id)

    def test_update_instance_by_id(self):
        ctx, cluster = self._make_sample()
        ng_id = cluster.node_groups[0].id
        inst_id = self.api.instance_add(ctx, ng_id, SAMPLE_INSTANCE)

        self.api.instance_update(ctx, inst_id, {'instance_name': 'tst123'})
        cluster = self.api.cluster_get(ctx, cluster.id)

        ng = gu.get_by_id(cluster.node_groups, ng_id)
        self.assertEqual('tst123', ng.instances[0].instance_name)

    def test_instance_volume_ops(self):
        ctx, cluster = self._make_sample()
        ng_id = cluster.node_groups[0].id
        inst_id = self.api.instance_add(ctx, ng_id, SAMPLE_INSTANCE)

        self.api.append_volume(ctx, inst_id, 0)
        self.api.append_volume(ctx, inst_id, 1)

        cluster = self.api.cluster_get(ctx, cluster.id)
        ng = gu.get_by_id(cluster.node_groups, ng_id)

        self.assertEqual(2, len(gu.get_by_id(ng.instances, inst_id).volumes))

        self.api.remove_volume(ctx, inst_id, 0)

        cluster = self.api.cluster_get(ctx, cluster.id)
        ng = gu.get_by_id(cluster.node_groups, ng_id)

        self.assertEqual(1, len(gu.get_by_id(ng.instances, inst_id).volumes))

    def _get_events(self, ctx, cluster_id, step_id=None):
        cluster = self.api.cluster_get(ctx, cluster_id, show_progress=True)
        events = []
        for step in cluster.provision_progress:
            if step_id == step['id']:
                return step['events']
            else:
                events += step['events']
        if step_id:
            return events
        else:
            return []

    def test_events_ops(self):
        ctx, cluster = self._make_sample()

        st_name = "some_name"
        st_type = "some_type"
        st_info = "some_info"

        # test provision step creation

        step_id = self.api.cluster_provision_step_add(ctx, cluster.id, {
            'step_name': st_name,
            'step_type': st_type,
        })

        ncluster = self.api.cluster_get(ctx, cluster.id)
        self.assertEqual(1, len(ncluster['provision_progress']))
        provision_step = ncluster['provision_progress'][0]

        self.assertEqual(st_name, provision_step['step_name'])
        self.assertEqual(st_type, provision_step['step_type'])
        self.assertEqual(cluster.id, provision_step['cluster_id'])

        # test adding event to step and getting events from step

        self.api.cluster_event_add(ctx, step_id, {
            'node_group_id': 'node_group_id',
            'instance_id': 'instance_id',
            'instance_name': st_name,
            'event_info': st_info,
            'successful': True
        })

        events = self._get_events(ctx, cluster.id, step_id)
        self.assertEqual(1, len(events))
        self.assertEqual(st_name, events[0].instance_name)
        self.assertTrue(events[0].successful)
        self.assertEqual(st_info, events[0].event_info)

        self.api.cluster_destroy(ctx, cluster.id)

        with testtools.ExpectedException(exceptions.NotFoundException):
            self._get_events(ctx, cluster.id, step_id)

    def test_job_main_name(self):
        ctx = context.ctx()
        job_binary = self.api.job_binary_create(ctx, SAMPLE_JOB_BINARY)
        job_binary_id = job_binary["id"]

        job_values = copy.copy(SAMPLE_JOB)
        job_values["mains"] = [job_binary_id]
        job = self.api.job_create(ctx, job_values)
        name = self.api.job_main_name(ctx, job)

        self.assertEqual(SAMPLE_JOB_BINARY["name"], name)

    def test_job_no_main_name(self):
        ctx = context.ctx()
        job = self.api.job_create(ctx, SAMPLE_JOB)
        name = self.api.job_main_name(ctx, job)

        self.assertIsNone(name)

    def test_job_libs_names(self):
        ctx = context.ctx()
        job_binary = self.api.job_binary_create(ctx, SAMPLE_JOB_BINARY)
        job_binary_id_0 = job_binary["id"]

        jb_1_values = copy.copy(SAMPLE_JOB_BINARY)
        jb_1_values["name"] = "some_other_name"
        job_binary = self.api.job_binary_create(ctx, jb_1_values)
        job_binary_id_1 = job_binary["id"]

        job_values = copy.copy(SAMPLE_JOB)
        job_values["libs"] = [job_binary_id_0, job_binary_id_1]
        job = self.api.job_create(ctx, job_values)

        names = self.api.job_lib_names(ctx, job)

        self.assertEqual(["job_binary_test", "some_other_name"], names)



# Copyright (c) 2014 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import os

from oslo_serialization import jsonutils as json
from oslo_utils import uuidutils
import testtools

from sahara.service.validations.edp import data_source_schema
from sahara.service.validations.edp import job_binary_schema
from sahara.service.validations.edp import job_execution_schema
from sahara.service.validations.edp import job_schema
from sahara.utils import api_validator


class TestJSONApiExamplesV11(testtools.TestCase):

    EXAMPLES_PATH = 'etc/edp-examples/json-api-examples/v1.1/%s'

    def test_data_sources(self):
        schema = data_source_schema.DATA_SOURCE_SCHEMA
        path = self.EXAMPLES_PATH % 'data-sources'
        formatter = self._formatter()
        self._test(schema, path, formatter)

    def test_job_binaries(self):
        schema = job_binary_schema.JOB_BINARY_SCHEMA
        path = self.EXAMPLES_PATH % 'job-binaries'
        formatter = self._formatter("job_binary_internal_id",
                                    "script_binary_internal_id",
                                    "text_binary_internal_id")
        self._test(schema, path, formatter)

    def test_jobs(self):
        schema = job_schema.JOB_SCHEMA
        path = self.EXAMPLES_PATH % 'jobs'
        formatter = self._formatter("job_binary_id",
                                    "udf_binary_id",
                                    "script_binary_id",
                                    "text_binary_id")
        self._test(schema, path, formatter)

    def test_job_executions(self):
        schema = job_execution_schema.JOB_EXEC_SCHEMA
        path = self.EXAMPLES_PATH % 'job-executions'
        formatter = self._formatter("cluster_id",
                                    "input_source_id",
                                    "output_source_id")
        self._test(schema, path, formatter)

    def _test(self, schema, path, formatter):
        validator = api_validator.ApiValidator(schema)
        for filename in self._files_in_path(path):
            file_path = '/'.join((path, filename))
            with open(file_path, 'r') as payload:
                payload = payload.read() % formatter
                payload = json.loads(payload)
                validator.validate(payload)

    def _files_in_path(self, path):
        all_files = (files for (path, directories, files) in os.walk(path))
        return itertools.chain(*all_files)

    def _formatter(self, *variables):
        return {variable: uuidutils.generate_uuid() for variable in variables}



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from oslo_config import cfg
from oslo_utils import excutils
from oslo_utils import timeutils
import six

from sahara import conductor as c
from sahara.conductor import resource
from sahara import context
from sahara.utils import cluster as cluster_utils

conductor = c.API
CONF = cfg.CONF

event_log_opts = [
    cfg.BoolOpt('disable_event_log',
                default=False,
                help="Disables event log feature.")
]


CONF.register_opts(event_log_opts)


def add_successful_event(instance):
    if CONF.disable_event_log:
        return

    cluster_id = instance.cluster_id
    step_id = get_current_provisioning_step(cluster_id)
    if step_id:
        conductor.cluster_event_add(context.ctx(), step_id, {
            'successful': True,
            'node_group_id': instance.node_group_id,
            'instance_id': instance.instance_id,
            'instance_name': instance.instance_name,
            'event_info': None,
        })


def add_fail_event(instance, exception):
    if CONF.disable_event_log:
        return

    cluster_id = instance.cluster_id
    step_id = get_current_provisioning_step(cluster_id)
    event_info = six.text_type(exception)

    if step_id:
        conductor.cluster_event_add(context.ctx(), step_id, {
            'successful': False,
            'node_group_id': instance.node_group_id,
            'instance_id': instance.instance_id,
            'instance_name': instance.instance_name,
            'event_info': event_info,
        })


def add_provisioning_step(cluster_id, step_name, total):
    if (CONF.disable_event_log or
            not cluster_utils.check_cluster_exists(cluster_id)):
        return

    prev_step = get_current_provisioning_step(cluster_id)
    if prev_step:
        conductor.cluster_provision_step_update(context.ctx(), prev_step)

    step_type = context.ctx().current_instance_info.step_type
    new_step = conductor.cluster_provision_step_add(
        context.ctx(), cluster_id, {
            'step_name': step_name,
            'step_type': step_type,
            'total': total,
            'started_at': timeutils.utcnow(),
        })
    context.current().current_instance_info.step_id = new_step
    return new_step


def get_current_provisioning_step(cluster_id):
    if (CONF.disable_event_log or
            not cluster_utils.check_cluster_exists(cluster_id)):
        return None
    current_instance_info = context.ctx().current_instance_info
    return current_instance_info.step_id


def event_wrapper(mark_successful_on_exit, **spec):
    """"General event-log wrapper

    :param mark_successful_on_exit: should we send success event
    after execution of function

    :param spec: extra specification
    :parameter step: provisioning step name (only for provisioning
    steps with only one event)
    :parameter param: tuple (name, pos) with parameter specification,
    where 'name' is the name of the parameter of function, 'pos' is the
    position of the parameter of function. This parameter is used to
    extract info about Instance or Cluster.
    """

    def decorator(func):
        @functools.wraps(func)
        def handler(*args, **kwargs):
            if CONF.disable_event_log:
                return func(*args, **kwargs)
            step_name = spec.get('step', None)
            instance = _find_in_args(spec, *args, **kwargs)
            cluster_id = instance.cluster_id

            if not cluster_utils.check_cluster_exists(cluster_id):
                return func(*args, **kwargs)

            if step_name:
                # It's single process, let's add provisioning step here
                add_provisioning_step(cluster_id, step_name, 1)

            try:
                value = func(*args, **kwargs)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    add_fail_event(instance, e)

            if mark_successful_on_exit:
                add_successful_event(instance)

            return value
        return handler
    return decorator


def _get_info_from_instance(arg):
    if isinstance(arg, resource.InstanceResource):
        return arg
    return None


def _get_info_from_cluster(arg):
    if isinstance(arg, resource.ClusterResource):
        return context.InstanceInfo(arg.id)
    return None


def _get_event_info(arg):
    try:
        return arg.get_event_info()
    except AttributeError:
        return None


def _get_info_from_obj(arg):
    functions = [_get_info_from_instance, _get_info_from_cluster,
                 _get_event_info]

    for func in functions:
        value = func(arg)
        if value:
            return value
    return None


def _find_in_args(spec, *args, **kwargs):
    param_values = spec.get('param', None)

    if param_values:
        p_name, p_pos = param_values
        obj = kwargs.get(p_name, None)
        if obj:
            return _get_info_from_obj(obj)
        return _get_info_from_obj(args[p_pos])

    # If param is not specified, let's search instance in args

    for arg in args:
        val = _get_info_from_instance(arg)
        if val:
            return val

    for arg in kwargs.values():
        val = _get_info_from_instance(arg)
        if val:
            return val

    # If instance not found in args, let's get instance info from context

    return context.ctx().current_instance_info



# Copyright (c) 2014 OpenStack Foundation
# Copyright (c) 2015 ISPRAS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from oslo_config import cfg
from oslo_utils import uuidutils

from sahara import conductor as c
from sahara import context
from sahara import exceptions as e
from sahara.i18n import _
from sahara.service.castellan import utils as key_manager
from sahara.service.edp import base_engine
from sahara.service.edp.job_binaries import manager as jb_manager
from sahara.service.edp import job_utils
from sahara.service.edp import s3_common
from sahara.service.validations.edp import job_execution as j
from sahara.swift import swift_helper as sw
from sahara.swift import utils as su
from sahara.utils import cluster as c_u
from sahara.utils import edp
from sahara.utils import files
from sahara.utils import remote
from sahara.utils import xmlutils

conductor = c.API
CONF = cfg.CONF


class SparkJobEngine(base_engine.JobEngine):
    def __init__(self, cluster):
        self.cluster = cluster
        # We'll always run the driver program on the master
        self.master = None
        # These parameters depend on engine that is used
        self.plugin_params = {"master": "",
                              "spark-user": "",
                              "deploy-mode": "",
                              "spark-submit": "",
                              "driver-class-path": "",
                              }

    def _get_pid_and_inst_id(self, job_id):
        try:
            pid, inst_id = job_id.split("@", 1)
            if pid and inst_id:
                return (pid, inst_id)
        except Exception:
            pass
        return "", ""

    def _get_instance_if_running(self, job_execution):
        pid, inst_id = self._get_pid_and_inst_id(job_execution.engine_job_id)
        if not pid or not inst_id or (
           job_execution.info['status'] in edp.JOB_STATUSES_TERMINATED):
            return None, None
        # TODO(tmckay): well, if there is a list index out of range
        # error here it probably means that the instance is gone. If we
        # have a job execution that is not terminated, and the instance
        # is gone, we should probably change the status somehow.
        # For now, do nothing.
        try:
            instance = c_u.get_instances(self.cluster, [inst_id])[0]
        except Exception:
            instance = None
        return pid, instance

    def _get_result_file(self, r, job_execution):
        result = os.path.join(job_execution.extra['spark-path'], "result")
        return r.execute_command("cat %s" % result,
                                 raise_when_error=False)

    def _check_pid(self, r, pid):
        ret, stdout = r.execute_command("ps hp %s" % pid,
                                        raise_when_error=False)
        return ret

    def _get_job_status_from_remote(self, r, pid, job_execution):
        # If the pid is there, it's still running
        if self._check_pid(r, pid) == 0:
            return {"status": edp.JOB_STATUS_RUNNING}

        # The process ended. Look in the result file to get the exit status
        ret, stdout = self._get_result_file(r, job_execution)
        if ret == 0:
            exit_status = stdout.strip()
            if exit_status == "0":
                return {"status": edp.JOB_STATUS_SUCCEEDED}
            # SIGINT will yield either -2 or 130
            elif exit_status in ["-2", "130"]:
                return {"status": edp.JOB_STATUS_KILLED}

        # Well, process is done and result is missing or unexpected
        return {"status": edp.JOB_STATUS_DONEWITHERROR}

    def _job_script(self, python_version):
        path = "service/edp/resources/launch_command.py"
        return files.get_file_text(path).replace(
            '{{PYTHON_VERSION}}', python_version)

    def _upload_wrapper_xml(self, where, job_dir, job_configs):
        xml_name = 'spark.xml'
        proxy_configs = job_configs.get('proxy_configs')
        configs = {}
        cfgs = job_configs.get('configs', {})
        if proxy_configs:
            configs[sw.HADOOP_SWIFT_USERNAME] = proxy_configs.get(
                'proxy_username')
            configs[sw.HADOOP_SWIFT_PASSWORD] = key_manager.get_secret(
                proxy_configs.get('proxy_password'))
            configs[sw.HADOOP_SWIFT_TRUST_ID] = proxy_configs.get(
                'proxy_trust_id')
            configs[sw.HADOOP_SWIFT_DOMAIN_NAME] = CONF.proxy_user_domain_name
        else:
            targets = [sw.HADOOP_SWIFT_USERNAME]
            configs = {k: cfgs[k] for k in targets if k in cfgs}
            if sw.HADOOP_SWIFT_PASSWORD in cfgs:
                configs[sw.HADOOP_SWIFT_PASSWORD] = (
                    key_manager.get_secret(cfgs[sw.HADOOP_SWIFT_PASSWORD])
                )

        for s3_cfg_key in s3_common.S3_DS_CONFIGS:
            if s3_cfg_key in cfgs:
                if s3_cfg_key == s3_common.S3_SECRET_KEY_CONFIG:
                    configs[s3_cfg_key] = (
                        key_manager.get_secret(cfgs[s3_cfg_key])
                    )
                else:
                    configs[s3_cfg_key] = cfgs[s3_cfg_key]

        content = xmlutils.create_hadoop_xml(configs)
        with remote.get_remote(where) as r:
            dst = os.path.join(job_dir, xml_name)
            r.write_file_to(dst, content)
        return xml_name

    def _prepare_job_binaries(self, job_binaries, r):
        for jb in job_binaries:
            jb_manager.JOB_BINARIES.get_job_binary_by_url(jb.url). \
                prepare_cluster(jb, remote=r)

    def _upload_job_files(self, where, job_dir, job, job_configs):

        def upload(r, dir, job_file, proxy_configs):
            path = jb_manager.JOB_BINARIES. \
                get_job_binary_by_url(job_file.url). \
                copy_binary_to_cluster(job_file,
                                       proxy_configs=proxy_configs,
                                       remote=r, context=context.ctx())
            return path

        def upload_builtin(r, dir, builtin):
            dst = os.path.join(dir, builtin['name'])
            r.write_file_to(dst, builtin['raw'])
            return dst

        builtin_libs = []
        if edp.is_adapt_spark_for_swift_enabled(
                job_configs.get('configs', {})):
            path = 'service/edp/resources/edp-spark-wrapper.jar'
            name = 'builtin-%s.jar' % uuidutils.generate_uuid()
            builtin_libs = [{'raw': files.try_get_file_text(path),
                             'name': name}]

        uploaded_paths = []
        builtin_paths = []
        with remote.get_remote(where) as r:
            mains = list(job.mains) if job.mains else []
            libs = list(job.libs) if job.libs else []

            job_binaries = mains + libs
            self._prepare_job_binaries(job_binaries, r)

            for job_file in job_binaries:
                uploaded_paths.append(
                    upload(r, job_dir, job_file,
                           job_configs.get('proxy_configs')))

            for builtin in builtin_libs:
                builtin_paths.append(
                    upload_builtin(r, job_dir, builtin))

        return uploaded_paths, builtin_paths

    def _check_driver_class_path(self, job_configs, param_dict, wf_dir):
        overridden = edp.spark_driver_classpath(
            job_configs.get('configs', {}))
        if overridden:
            param_dict['driver-class-path'] = (
                " --driver-class-path " + overridden)
            return
        if not param_dict.get('wrapper_jar'):
            # no need in driver classpath if swift as datasource is not used
            param_dict['driver-class-path'] = ""
            return
        cp = param_dict['driver-class-path'] or ""
        if param_dict['deploy-mode'] == 'client' and not (
                cp.startswith(":") or cp.endswith(":")):
            cp += ":" + wf_dir
        param_dict['driver-class-path'] = " --driver-class-path " + cp

    def cancel_job(self, job_execution):
        pid, instance = self._get_instance_if_running(job_execution)
        if instance is not None:
            with remote.get_remote(instance) as r:
                ret, stdout = r.execute_command("kill -SIGINT %s" % pid,
                                                raise_when_error=False)
                if ret == 0:
                    # We had some effect, check the status
                    return self._get_job_status_from_remote(r,
                                                            pid, job_execution)

    def get_job_status(self, job_execution):
        pid, instance = self._get_instance_if_running(job_execution)
        if instance is not None:
            with remote.get_remote(instance) as r:
                return self._get_job_status_from_remote(r, pid, job_execution)

    def _build_command(self, wf_dir, paths, builtin_paths,
                       updated_job_configs):

        indep_params = {}

        # TODO(tmckay): for now, paths[0] is always assumed to be the app
        # jar and we generate paths in order (mains, then libs).
        # When we have a Spark job type, we can require a "main" and set
        # the app jar explicitly to be "main"
        indep_params["app_jar"] = paths.pop(0)
        indep_params["job_class"] = (
            updated_job_configs["configs"]["edp.java.main_class"])
        if self.plugin_params.get('drivers-to-jars', None):
            paths.extend(self.plugin_params['drivers-to-jars'])
        # If we uploaded builtins then we are using a wrapper jar. It will
        # be the first one on the builtin list and the original app_jar needs
        # to be added to the  'additional' jars
        if builtin_paths:
            indep_params["wrapper_jar"] = builtin_paths.pop(0)
            indep_params["wrapper_class"] = (
                'org.openstack.sahara.edp.SparkWrapper')
            wrapper_xml = self._upload_wrapper_xml(self.master,
                                                   wf_dir,
                                                   updated_job_configs)
            indep_params["wrapper_args"] = "%s %s" % (
                wrapper_xml, indep_params["job_class"])

            indep_params["addnl_files"] = wrapper_xml

            indep_params["addnl_jars"] = ",".join(
                [indep_params["wrapper_jar"]] + paths + builtin_paths)

        else:
            indep_params["addnl_jars"] = ",".join(paths)

        # All additional jars are passed with the --jars option
        if indep_params["addnl_jars"]:
            indep_params["addnl_jars"] = (
                " --jars " + indep_params["addnl_jars"])

        # Launch the spark job using spark-submit and deploy_mode = client
        # TODO(tmckay): we need to clean up wf_dirs on long running clusters
        # TODO(tmckay): probably allow for general options to spark-submit
        indep_params["args"] = updated_job_configs.get('args', [])
        indep_params["args"] = " ".join([su.inject_swift_url_suffix(arg)
                                         for arg in indep_params["args"]])
        if indep_params.get("args"):
            indep_params["args"] = (" " + indep_params["args"])

        mutual_dict = self.plugin_params.copy()
        mutual_dict.update(indep_params)

        # Handle driver classpath. Because of the way the hadoop
        # configuration is handled in the wrapper class, using
        # wrapper_xml, the working directory must be on the classpath
        self._check_driver_class_path(updated_job_configs, mutual_dict, wf_dir)

        if mutual_dict.get("wrapper_jar"):
            # Substrings which may be empty have spaces
            # embedded if they are non-empty
            cmd = (
                '%(spark-user)s%(spark-submit)s%(driver-class-path)s'
                ' --files %(addnl_files)s'
                ' --class %(wrapper_class)s%(addnl_jars)s'
                ' --master %(master)s'
                ' --deploy-mode %(deploy-mode)s'
                ' %(app_jar)s %(wrapper_args)s%(args)s') % dict(
                mutual_dict)
        else:
            cmd = (
                '%(spark-user)s%(spark-submit)s%(driver-class-path)s'
                ' --class %(job_class)s%(addnl_jars)s'
                ' --master %(master)s'
                ' --deploy-mode %(deploy-mode)s'
                ' %(app_jar)s%(args)s') % dict(
                mutual_dict)

        return cmd

    def run_job(self, job_execution):
        ctx = context.ctx()
        job = conductor.job_get(ctx, job_execution.job_id)
        # This will be a dictionary of tuples, (native_url, runtime_url)
        # keyed by data_source id
        data_source_urls = {}
        additional_sources, updated_job_configs = (
            job_utils.resolve_data_source_references(job_execution.job_configs,
                                                     job_execution.id,
                                                     data_source_urls,
                                                     self.cluster)
        )

        job_execution = conductor.job_execution_update(
            ctx, job_execution,
            {"data_source_urls": job_utils.to_url_dict(data_source_urls)})

        # Now that we've recorded the native urls, we can switch to the
        # runtime urls
        data_source_urls = job_utils.to_url_dict(data_source_urls,
                                                 runtime=True)

        job_utils.prepare_cluster_for_ds(additional_sources,
                                         self.cluster, updated_job_configs,
                                         data_source_urls)

        # It is needed in case we are working with Spark plugin
        self.plugin_params['master'] = (
            self.plugin_params['master'] % {'host': self.master.hostname()})

        # TODO(tmckay): wf_dir should probably be configurable.
        # The only requirement is that the dir is writable by the image user
        wf_dir = job_utils.create_workflow_dir(self.master, '/tmp/spark-edp',
                                               job, job_execution.id, "700")
        paths, builtin_paths = self._upload_job_files(
            self.master, wf_dir, job, updated_job_configs)

        # We can shorten the paths in this case since we'll run out of wf_dir
        paths = [os.path.basename(p) if p.startswith(wf_dir) else p
                 for p in paths]
        builtin_paths = [os.path.basename(p) for p in builtin_paths]

        cmd = self._build_command(wf_dir, paths, builtin_paths,
                                  updated_job_configs)

        job_execution = conductor.job_execution_get(ctx, job_execution.id)
        if job_execution.info['status'] == edp.JOB_STATUS_TOBEKILLED:
            return (None, edp.JOB_STATUS_KILLED, None)

        # If an exception is raised here, the job_manager will mark
        # the job failed and log the exception
        # The redirects of stdout and stderr will preserve output in the wf_dir
        with remote.get_remote(self.master) as r:
            # Upload the command launch script
            launch = os.path.join(wf_dir, "launch_command")
            python_version = r.get_python_version()
            r.write_file_to(launch, self._job_script(python_version))
            r.execute_command("chmod u+rwx,g+rx,o+rx %s" % wf_dir)
            r.execute_command("chmod +x %s" % launch)
            ret, stdout = r.execute_command(
                "cd %s; ./launch_command %s > /dev/null 2>&1 & echo $!"
                % (wf_dir, cmd))

        if ret == 0:
            # Success, we'll add the wf_dir in job_execution.extra and store
            # pid@instance_id as the job id

            # We know the job is running so return "RUNNING"
            return (stdout.strip() + "@" + self.master.id,
                    edp.JOB_STATUS_RUNNING,
                    {'spark-path': wf_dir})

        # Hmm, no execption but something failed.
        # Since we're using backgrounding with redirect, this is unlikely.
        raise e.EDPError(_("Spark job execution failed. Exit status = "
                           "%(status)s, stdout = %(stdout)s") %
                         {'status': ret, 'stdout': stdout})

    def run_scheduled_job(self, job_execution):
        raise e.NotImplementedException(_("Currently Spark engine does not"
                                          " support scheduled EDP jobs"))

    def validate_job_execution(self, cluster, job, data):
        j.check_main_class_present(data, job)

    @staticmethod
    def get_possible_job_config(job_type):
        return {'job_config': {'configs': [], 'args': []}}

    @staticmethod
    def get_supported_job_types():
        return [edp.JOB_TYPE_SPARK]


class SparkShellJobEngine(SparkJobEngine):
    def _build_command(self, wf_dir, paths, builtin_paths,
                       updated_job_configs):
        main_script = paths.pop(0)
        args = " ".join(updated_job_configs.get('args', []))

        env_params = ""
        params = updated_job_configs.get('params', {})
        for key, value in params.items():
            env_params += "{key}={value} ".format(key=key, value=value)

        cmd = ("{env_params}{cmd} {main_script} {args}".format(
            cmd='/bin/sh', main_script=main_script, env_params=env_params,
            args=args))

        return cmd

    def validate_job_execution(self, cluster, job, data):
        # Shell job doesn't require any special validation
        pass

    @staticmethod
    def get_possible_job_config(job_type):
        return {'job_config': {'configs': {}, 'args': [], 'params': {}}}

    @staticmethod
    def get_supported_job_types():
        return [edp.JOB_TYPE_SHELL]



# Copyright (c) 2017 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import mock
from oslo_utils import uuidutils
import testtools

import sahara.exceptions as ex
from sahara.service.edp.data_sources.swift.implementation import SwiftType
from sahara.service.edp import job_utils
from sahara.swift import utils as su
from sahara.tests.unit import base
from sahara.tests.unit.service.edp import edp_test_utils as u
from sahara.utils.types import FrozenDict

SAMPLE_SWIFT_URL = "swift://1234/object"
SAMPLE_SWIFT_URL_WITH_SUFFIX = "swift://1234%s/object" % su.SWIFT_URL_SUFFIX


class TestSwiftTypeValidation(base.SaharaTestCase):
    def setUp(self):
        super(TestSwiftTypeValidation, self).setUp()
        self.s_type = SwiftType()

    @mock.patch('sahara.context.ctx')
    def test_prepare_cluster(self, ctx):

        ctx.return_value = 'dummy'

        ds_url = "swift://container/input"
        ds = u.create_data_source(ds_url,
                                  name="data_source",
                                  id=uuidutils.generate_uuid())

        job_configs = {
            'configs': {
                job_utils.DATA_SOURCE_SUBST_NAME: True,
                job_utils.DATA_SOURCE_SUBST_UUID: True
                }
            }

        old_configs = copy.deepcopy(job_configs)

        self.s_type.prepare_cluster(ds, u.create_cluster(),
                                    job_configs=job_configs)
        # Swift configs should be filled in since they were blank
        self.assertEqual(ds.credentials['user'],
                         job_configs['configs']
                         ['fs.swift.service.sahara.username'])
        self.assertEqual(ds.credentials['password'],
                         job_configs['configs']
                         ['fs.swift.service.sahara.password'])

        self.assertNotEqual(old_configs, job_configs)

        job_configs['configs'] = {'fs.swift.service.sahara.username': 'sam',
                                  'fs.swift.service.sahara.password': 'gamgee',
                                  job_utils.DATA_SOURCE_SUBST_NAME: False,
                                  job_utils.DATA_SOURCE_SUBST_UUID: True}
        old_configs = copy.deepcopy(job_configs)

        self.s_type.prepare_cluster(ds, u.create_cluster(),
                                    job_configs=job_configs)
        # Swift configs should not be overwritten
        self.assertEqual(old_configs['configs'], job_configs['configs'])

        job_configs['configs'] = {job_utils.DATA_SOURCE_SUBST_NAME: True,
                                  job_utils.DATA_SOURCE_SUBST_UUID: False}
        job_configs['proxy_configs'] = {'proxy_username': 'john',
                                        'proxy_password': 'smith',
                                        'proxy_trust_id': 'trustme'}
        old_configs = copy.deepcopy(job_configs)
        self.s_type.prepare_cluster(ds, u.create_cluster(),
                                    job_configs=job_configs)

        # Swift configs should be empty and proxy configs should be preserved
        self.assertEqual(old_configs['configs'], job_configs['configs'])
        self.assertEqual(old_configs['proxy_configs'],
                         job_configs['proxy_configs'])

        # If there's no configs do nothing
        job_configs['configs'] = None
        old_configs = copy.deepcopy(job_configs)

        self.s_type.prepare_cluster(ds, u.create_cluster(),
                                    job_configs=job_configs)
        self.assertEqual(old_configs, job_configs)

        # If it's a FrozenDict do nothing
        job_configs = {
            'configs': {
                job_utils.DATA_SOURCE_SUBST_NAME: True,
                job_utils.DATA_SOURCE_SUBST_UUID: True
                }
            }

        old_configs = copy.deepcopy(job_configs)
        job_configs = FrozenDict(job_configs)

        self.s_type.prepare_cluster(ds, u.create_cluster(),
                                    job_configs=job_configs)
        self.assertEqual(old_configs, job_configs)

    def test_swift_type_validation(self):
        data = {
            "name": "test_data_data_source",
            "url": SAMPLE_SWIFT_URL,
            "type": "swift",
            "credentials": {
                "user": "user",
                "password": "password"
            },
            "description": "long description"
        }
        self.s_type.validate(data)

    def test_swift_type_validation_missing_credentials(self):
        data = {
            "name": "test_data_data_source",
            "url": SAMPLE_SWIFT_URL,
            "type": "swift",
            "description": "long description"
        }
        with testtools.ExpectedException(ex.InvalidCredentials):
            self.s_type.validate(data)
        # proxy enabled should allow creation without credentials
        self.override_config('use_domain_for_proxy_users', True)
        self.s_type.validate(data)

    def test_swift_type_validation_credentials_missing_user(self):
        data = {
            "name": "test_data_data_source",
            "url": SAMPLE_SWIFT_URL,
            "type": "swift",
            "credentials": {
                "password": "password"
            },
            "description": "long description"
        }
        with testtools.ExpectedException(ex.InvalidCredentials):
            self.s_type.validate(data)
        # proxy enabled should allow creation without credentials
        self.override_config('use_domain_for_proxy_users', True)
        self.s_type.validate(data)

    def test_swift_type_validation_credentials_missing_password(self):
        data = {
            "name": "test_data_data_source",
            "url": SAMPLE_SWIFT_URL,
            "type": "swift",
            "credentials": {
                "user": "user",
            },
            "description": "long description"
        }
        with testtools.ExpectedException(ex.InvalidCredentials):
            self.s_type.validate(data)
        # proxy enabled should allow creation without credentials
        self.override_config('use_domain_for_proxy_users', True)
        self.s_type.validate(data)

    def test_swift_type_validation_wrong_schema(self):
        data = {
            "name": "test_data_data_source",
            "url": "swif://1234/object",
            "type": "swift",
            "description": "incorrect url schema"
        }
        with testtools.ExpectedException(ex.InvalidDataException):
            self.s_type.validate(data)

    def test_swift_type_validation_explicit_suffix(self):
        data = {
            "name": "test_data_data_source",
            "url": SAMPLE_SWIFT_URL_WITH_SUFFIX,
            "type": "swift",
            "description": "incorrect url schema",
            "credentials": {
                "user": "user",
                "password": "password"
            }
        }
        self.s_type.validate(data)

    def test_swift_type_validation_wrong_suffix(self):
        data = {
            "name": "test_data_data_source",
            "url": "swift://1234.suffix/object",
            "type": "swift",
            "description": "incorrect url schema"
        }
        with testtools.ExpectedException(ex.InvalidDataException):
            self.s_type.validate(data)

    def test_swift_type_validation_missing_object(self):
        data = {
            "name": "test_data_data_source",
            "url": "swift://1234/",
            "type": "swift",
            "description": "incorrect url schema"
        }
        with testtools.ExpectedException(ex.InvalidDataException):
            self.s_type.validate(data)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from oslo_log import log as logging

from sahara import conductor as cond
from sahara import context
from sahara import exceptions as ex
from sahara.i18n import _

conductor = cond.API
LOG = logging.getLogger(__name__)


STABLE = {
    'name': 'stable',
    'mutable': False,
    'description': "Indicates that plugin or its version are stable to be used"
}

DEPRECATED = {
    'name': 'deprecated',
    'mutable': False,
    'description': "Plugin or its version is deprecated and will be removed "
                   "in future releases. Please, consider to use another "
                   "plugin or its version."
}

ENABLED = {
    'name': 'enabled',
    'mutable': True,
    'description': "Plugin or its version is enabled and can be used by user."
}

HIDDEN = {
    'name': 'hidden',
    'mutable': True,
    'description': "Existence of plugin or its version is hidden, but "
                   "still can be used for cluster creation by CLI and "
                   "directly by client."
}

PLUGIN_LABELS_SCOPE = 'plugin_labels'
VERSION_LABELS_SCOPE = 'version_labels'
MUTABLE = 'mutable'

LABEL_OBJECT = {
    'type': 'object',
    'properties': {
        'status': {
            'type': 'boolean',
        }
    },
    "additionalProperties": False,
}


class LabelHandler(object):
    def __init__(self, loaded_plugins):
        self.plugins = loaded_plugins

    def get_plugin_update_validation_jsonschema(self):
        schema = {
            'type': 'object', "additionalProperties": False,
            'properties': {
                VERSION_LABELS_SCOPE: {
                    'type': 'object', 'additionalProperties': False,
                },
            },
        }
        ln = [label['name'] for label in self.get_labels()]
        labels_descr_object = {
            'type': 'object',
            "properties": {name: copy.deepcopy(LABEL_OBJECT) for name in ln},
            "additionalProperties": False
        }
        schema['properties'][PLUGIN_LABELS_SCOPE] = copy.deepcopy(
            labels_descr_object)
        all_versions = []
        for plugin_name in self.plugins.keys():
            plugin = self.plugins[plugin_name]
            all_versions.extend(plugin.get_versions())
        all_versions = set(all_versions)
        schema['properties'][VERSION_LABELS_SCOPE]['properties'] = {
            ver: copy.deepcopy(labels_descr_object) for ver in all_versions
        }
        return schema

    def get_default_label_details(self, plugin_name):
        plugin = self.plugins.get(plugin_name)
        return plugin.get_labels()

    def get_label_details(self, plugin_name):
        try:
            plugin = conductor.plugin_get(context.ctx(), plugin_name)
        except Exception:
            LOG.error("Unable to retrieve plugin data from database")
            plugin = None
        if not plugin:
            plugin = self.get_default_label_details(plugin_name)
        fields = ['name', 'id', 'updated_at', 'created_at']
        for field in fields:
            if field in plugin:
                del plugin[field]
        return plugin

    def get_label_full_details(self, plugin_name):
        return self.expand_data(self.get_label_details(plugin_name))

    def get_labels(self):
        return [HIDDEN, STABLE, ENABLED, DEPRECATED]

    def get_labels_map(self):
        return {
            label['name']: label for label in self.get_labels()
        }

    def expand_data(self, plugin):
        plugin_labels = plugin.get(PLUGIN_LABELS_SCOPE)
        labels_map = self.get_labels_map()
        for key in plugin_labels.keys():
            key_desc = labels_map.get(key)
            plugin_labels[key].update(key_desc)
            del plugin_labels[key]['name']

        for version in plugin.get(VERSION_LABELS_SCOPE):
            vers_labels = plugin.get(VERSION_LABELS_SCOPE).get(version)
            for key in vers_labels.keys():
                key_desc = labels_map.get(key)
                vers_labels[key].update(key_desc)
                del vers_labels[key]['name']

        return plugin

    def _validate_labels_update(self, default_data, update_values):
        for label in update_values.keys():
            if label not in default_data.keys():
                raise ex.InvalidDataException(
                    _("Label '%s' can't be updated because it's not "
                      "available for plugin or its version") % label)
            if not default_data[label][MUTABLE]:
                raise ex.InvalidDataException(
                    _("Label '%s' can't be updated because it's not "
                      "mutable") % label)

    def validate_plugin_update(self, plugin_name, values):
        plugin = self.plugins[plugin_name]
        # it's important to get full details since we have mutability
        default = self.get_label_full_details(plugin_name)
        if values.get(PLUGIN_LABELS_SCOPE):
            pl = values.get(PLUGIN_LABELS_SCOPE)
            self._validate_labels_update(default[PLUGIN_LABELS_SCOPE], pl)

        if values.get(VERSION_LABELS_SCOPE):
            vl = values.get(VERSION_LABELS_SCOPE)
            for version in vl.keys():
                if version not in plugin.get_versions():
                    raise ex.InvalidDataException(
                        _("Unknown plugin version '%(version)s' of "
                          "%(plugin)s") % {
                            'version': version, 'plugin': plugin_name})
                self._validate_labels_update(
                    default[VERSION_LABELS_SCOPE][version], vl[version])

    def update_plugin(self, plugin_name, values):
        ctx = context.ctx()
        current = self.get_label_details(plugin_name)
        if not conductor.plugin_get(ctx, plugin_name):
            current['name'] = plugin_name
            conductor.plugin_create(ctx, current)
            del current['name']

        if values.get(PLUGIN_LABELS_SCOPE):
            for label in values.get(PLUGIN_LABELS_SCOPE).keys():
                current[PLUGIN_LABELS_SCOPE][label].update(
                    values.get(PLUGIN_LABELS_SCOPE).get(label))
        else:
            del current[PLUGIN_LABELS_SCOPE]

        if values.get(VERSION_LABELS_SCOPE):
            vl = values.get(VERSION_LABELS_SCOPE)
            for version in vl.keys():
                for label in vl.get(version).keys():
                    current[VERSION_LABELS_SCOPE][version][label].update(
                        vl[version][label])
        else:
            del current[VERSION_LABELS_SCOPE]

        conductor.plugin_update(context.ctx(), plugin_name, current)

    def validate_plugin_labels(self, plugin_name, version):
        details = self.get_label_details(plugin_name)
        plb = details.get(PLUGIN_LABELS_SCOPE, {})
        if not plb.get('enabled', {}).get('status'):
            raise ex.InvalidReferenceException(
                _("Plugin %s is not enabled") % plugin_name)

        if plb.get('deprecated', {}).get('status', False):
            LOG.warning("Plugin %s is deprecated and can be removed in "
                        "the next release", plugin_name)

        vlb = details.get(VERSION_LABELS_SCOPE, {}).get(version, {})
        if not vlb.get('enabled', {}).get('status'):
            raise ex.InvalidReferenceException(
                _("Version %(version)s of plugin %(plugin)s is not enabled")
                % {'version': version, 'plugin': plugin_name})

        if vlb.get('deprecated', {}).get('status', False):
            LOG.warning("Using version %(version)s of plugin %(plugin)s is "
                        "deprecated and can removed in next release",
                        {'version': version, 'plugin': plugin_name})



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

import time

from oslo_utils import timeutils

from sahara import conductor as c
from sahara import context
from sahara import exceptions as ex
from sahara.i18n import _
from sahara.plugins import base as plugin_base
from sahara.service.edp import job_utils as j_u
from sahara.service.validations import acl
from sahara.service.validations import base as val_base
import sahara.service.validations.edp.base as b
import sahara.service.validations.edp.job_interface as j_i
from sahara.utils import cluster as c_u
from sahara.utils import edp


conductor = c.API


def _is_main_class_present(data):
    if data:
        val = data.get(
            'job_configs', {}).get(
                'configs', {}).get('edp.java.main_class', None)
        return val and isinstance(val, six.string_types)
    return False


def check_main_class_present(data, job):
    if not _is_main_class_present(data):
        raise ex.InvalidDataException(
            _('%s job must specify edp.java.main_class') % job.type)


def _is_topology_name_present(data):
    if data:
        val = data.get(
            'job_configs', {}).get(
            'configs', {}).get('topology_name', None)
        return val and isinstance(val, six.string_types)
    return False


def check_topology_name_present(data, job):
    if not _is_main_class_present(data):
        raise ex.InvalidDataException(
            _('%s job must specify topology_name') % job.type)


def _streaming_present(data):
    try:
        streaming = set(('edp.streaming.mapper',
                         'edp.streaming.reducer'))
        configs = set(data['job_configs']['configs'])
        return streaming.intersection(configs) == streaming
    except Exception:
        return False


def check_streaming_present(data, job):
    if not _streaming_present(data):
        raise ex.InvalidDataException(
            _("%s job must specify streaming mapper and reducer") % job.type)


def check_scheduled_job_execution_info(job_execution_info):
    start = job_execution_info.get('start', None)
    if start is None:
        raise ex.InvalidDataException(_(
            "Scheduled job must specify start time"))
    try:
        start = time.strptime(start, "%Y-%m-%d %H:%M:%S")
        start = timeutils.datetime.datetime.fromtimestamp(time.mktime(start))
    except Exception:
        raise ex.InvalidDataException(_("Invalid Time Format"))

    now_time = timeutils.utcnow()

    if timeutils.delta_seconds(now_time, start) < 0:
        raise ex.InvalidJobExecutionInfoException(_(
            "Job start time should be later than now"))


def check_job_execution(data, job_templates_id=None):
    ctx = context.ctx()
    job_execution_info = data.get('job_execution_info', {})

    cluster = conductor.cluster_get(ctx, data['cluster_id'])
    if not cluster:
        raise ex.InvalidReferenceException(
            _("Cluster with id '%s' doesn't exist") % data['cluster_id'])

    val_base.check_plugin_labels(cluster.plugin_name, cluster.hadoop_version)

    jt_err_msg = _("Job template with id '%s' doesn't exist")
    if job_templates_id is None:
        job = conductor.job_get(ctx, data['job_template_id'])
        if not job:
            raise ex.InvalidReferenceException(
                jt_err_msg % data['job_template_id'])
    else:
        job = conductor.job_get(ctx, job_templates_id)
        if not job:
            raise ex.InvalidReferenceException(
                jt_err_msg % job_templates_id)

    plugin = plugin_base.PLUGINS.get_plugin(cluster.plugin_name)
    edp_engine = plugin.get_edp_engine(cluster, job.type)
    if not edp_engine:
        raise ex.InvalidReferenceException(
            _("Cluster with id '%(cluster_id)s' doesn't support job type "
              "'%(job_type)s'") % {"cluster_id": cluster.id,
                                   "job_type": job.type})

    j_i.check_execution_interface(data, job)
    edp_engine.validate_job_execution(cluster, job, data)

    if 'job_execution_type' in job_execution_info:
        j_type = job_execution_info.get('job_execution_type', 'workflow')
        if j_type == 'scheduled':
            check_scheduled_job_execution_info(job_execution_info)


def check_data_sources(data, job):
    if not ('input_id' in data and 'output_id' in data):
        raise ex.InvalidDataException(_("%s job requires 'input_id' "
                                        "and 'output_id'") % job.type)

    b.check_data_source_exists(data['input_id'])
    b.check_data_source_exists(data['output_id'])

    b.check_data_sources_are_different(data['input_id'], data['output_id'])


def check_job_execution_cancel(job_id, **kwargs):
    ctx = context.current()
    je = conductor.job_execution_get(ctx, job_id)

    if je.tenant_id != ctx.tenant_id:
        raise ex.CancelingFailed(
            _("Job execution with id '%s' cannot be canceled "
              "because it wasn't created in this tenant")
            % job_id)

    if je.is_protected:
        raise ex.CancelingFailed(
            _("Job Execution with id '%s' cannot be canceled "
              "because it's marked as protected") % job_id)


def check_job_execution_delete(job_id, **kwargs):
    ctx = context.current()
    je = conductor.job_execution_get(ctx, job_id)

    acl.check_tenant_for_delete(ctx, je)
    acl.check_protected_from_delete(je)


def check_job_execution_update(job_id, data, **kwargs):
    ctx = context.current()
    je = conductor.job_execution_get(ctx, job_id)

    acl.check_tenant_for_update(ctx, je)
    acl.check_protected_from_update(je, data)


def check_job_status_update(job_id, data):
    ctx = context.ctx()
    job_execution = conductor.job_execution_get(ctx, job_id)
    # check we are updating status
    if 'info' in data:
        if len(data) != 1:
            raise ex.InvalidJobStatus(_("Invalid status parameter"))
    cluster = conductor.cluster_get(ctx, job_execution.cluster_id)
    if cluster is None or cluster.status != c_u.CLUSTER_STATUS_ACTIVE:
        raise ex.InvalidDataException(
            _("Suspending operation can not be performed on an inactive or "
              "non-existent cluster"))
    job_templates_id = conductor.job_execution_get(ctx, job_id).job_id
    job_type = conductor.job_get(ctx, job_templates_id).type
    engine = j_u.get_plugin(cluster).get_edp_engine(cluster, job_type)
    if 'info' in data:
        if data['info']['status'] == edp.JOB_ACTION_SUSPEND:
            if not engine.does_engine_implement('suspend_job'):
                raise ex.InvalidReferenceException(
                    _("Engine doesn't support suspending job feature"))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
import six

from sahara import conductor as c
from sahara import context
from sahara import exceptions as ex
from sahara.plugins import base as plugin_base
from sahara.service import api
from sahara.service.edp.binary_retrievers import dispatch
from sahara.service.edp import job_manager as manager
from sahara.utils import edp
from sahara.utils import proxy as p


conductor = c.API
LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def get_job_types(**kwargs):
    # Return a dictionary of all the job types that can be run
    # by this instance of Sahara. For each job type, the value
    # will be a list of plugins that support the job type. For
    # each plugin, include a dictionary of the versions that
    # support the job type.

    # All entries in kwargs are expected to have list values
    hints = kwargs.get("hints", ["false"])[0].lower() == "true"

    plugin_names = kwargs.get("plugin", [])
    all_plugins = plugin_base.PLUGINS.get_plugins()
    if plugin_names:
        plugins = filter(lambda x: x.name in plugin_names, all_plugins)
    else:
        plugins = all_plugins

    job_types = kwargs.get("type", edp.JOB_TYPES_ALL)
    versions = kwargs.get("version", [])

    res = []

    for job_type in job_types:
        # All job types supported by all versions of the plugin.
        # This is a dictionary where keys are plugin version
        # strings and values are lists of job types
        job_entry = {"name": job_type,
                     "plugins": []}

        for plugin in plugins:
            types_for_plugin = plugin.get_edp_job_types(versions)

            # dict returns a new object so we are not modifying the plugin
            p = plugin.dict

            # Find only the versions of this plugin that support the job.
            # Additionally, instead of a list we want a dictionary of
            # plugin versions with corresponding config hints
            p["versions"] = {}

            for version, supported_types in six.iteritems(types_for_plugin):
                if job_type in supported_types:
                    if hints:
                        config_hints = plugin.get_edp_config_hints(job_type,
                                                                   version)
                    else:
                        config_hints = {}
                    p["versions"][version] = config_hints

            # If we found at least one version of the plugin that
            # supports the job type, add the plugin to the result
            if p["versions"]:
                job_entry["plugins"].append(p)

        if job_entry["plugins"]:
            res.append(job_entry)
    return res


def get_job_config_hints(job_type):
    return manager.get_job_config_hints(job_type)


def execute_job(job_id, data):
    # Elements common to all job types
    cluster_id = data['cluster_id']
    configs = data.get('job_configs', {})
    interface = data.get('interface', {})

    # Not in Java job types but present for all others
    input_id = data.get('input_id', None)
    output_id = data.get('output_id', None)

    # Since we will use a unified class in the database, we pass
    # a superset for all job types
    # example configs['start'] = '2015-05-12T08:55Z' frequency = 5 mins
    # the job will starts from 2015-05-12T08:55Z, runs every 5 mins

    job_execution_info = data.get('job_execution_info', {})

    configs['job_execution_info'] = job_execution_info

    job_ex_dict = {'input_id': input_id, 'output_id': output_id,
                   'job_id': job_id, 'cluster_id': cluster_id,
                   'info': {'status': edp.JOB_STATUS_PENDING},
                   'job_configs': configs, 'extra': {},
                   'interface': interface}
    job_execution = conductor.job_execution_create(context.ctx(), job_ex_dict)
    context.set_current_job_execution_id(job_execution.id)

    # check to use proxy user
    if p.job_execution_requires_proxy_user(job_execution):
        try:
            p.create_proxy_user_for_job_execution(job_execution)
        except ex.SaharaException as e:
            LOG.error("Can't run job execution. "
                      "(Reasons: {reason})".format(reason=e))
            conductor.job_execution_destroy(context.ctx(), job_execution)
            raise

    api.OPS.run_edp_job(job_execution.id)

    return job_execution


def get_job_execution_status(id):
    return manager.get_job_status(id)


def job_execution_list(**kwargs):
    return conductor.job_execution_get_all(context.ctx(),
                                           regex_search=True, **kwargs)


def get_job_execution(id):
    return conductor.job_execution_get(context.ctx(), id)


def cancel_job_execution(id):
    context.set_current_job_execution_id(id)
    job_execution = conductor.job_execution_get(context.ctx(), id)
    api.OPS.cancel_job_execution(id)

    return job_execution


def update_job_execution(id, values):
    _update_status(values.pop("info", None), id)
    return conductor.job_execution_update(context.ctx(), id, values)


def _update_status(info, id):
    if info:
        status = info.get("status", None)
        if status == edp.JOB_ACTION_SUSPEND:
            api.OPS.job_execution_suspend(id)


def delete_job_execution(id):
    context.set_current_job_execution_id(id)
    api.OPS.delete_job_execution(id)


def get_data_sources(**kwargs):
    return conductor.data_source_get_all(context.ctx(),
                                         regex_search=True, **kwargs)


def get_data_source(id):
    return conductor.data_source_get(context.ctx(), id)


def delete_data_source(id):
    conductor.data_source_destroy(context.ctx(), id)


def register_data_source(values):
    return conductor.data_source_create(context.ctx(), values)


def data_source_update(id, values):
    return conductor.data_source_update(context.ctx(), id, values)


def get_jobs(**kwargs):
    return conductor.job_get_all(context.ctx(), regex_search=True, **kwargs)


def get_job(id):
    return conductor.job_get(context.ctx(), id)


def create_job(values):
    return conductor.job_create(context.ctx(), values)


def update_job(id, values):
    return conductor.job_update(context.ctx(), id, values)


def delete_job(job_id):
    return conductor.job_destroy(context.ctx(), job_id)


def create_job_binary(values):
    return conductor.job_binary_create(context.ctx(), values)


def get_job_binaries(**kwargs):
    return conductor.job_binary_get_all(context.ctx(),
                                        regex_search=True, **kwargs)


def get_job_binary(id):
    return conductor.job_binary_get(context.ctx(), id)


def update_job_binary(id, values):
    return conductor.job_binary_update(context.ctx(), id, values)


def delete_job_binary(id):
    conductor.job_binary_destroy(context.ctx(), id)


def create_job_binary_internal(values):
    return conductor.job_binary_internal_create(context.ctx(), values)


def get_job_binary_internals(**kwargs):
    return conductor.job_binary_internal_get_all(context.ctx(),
                                                 regex_search=True, **kwargs)


def get_job_binary_internal(id):
    return conductor.job_binary_internal_get(context.ctx(), id)


def delete_job_binary_internal(id):
    conductor.job_binary_internal_destroy(context.ctx(), id)


def get_job_binary_internal_data(id):
    return conductor.job_binary_internal_get_raw_data(context.ctx(), id)


def update_job_binary_internal(id, values):
    return conductor.job_binary_internal_update(context.ctx(), id, values)


def get_job_binary_data(id):
    job_binary = conductor.job_binary_get(context.ctx(), id)
    return dispatch.get_raw_binary(job_binary, with_context=True)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
import six

import sahara.exceptions as e
from sahara.i18n import _
from sahara.service.edp.job_binaries import manager as jb_manager
from sahara.utils import api_validator as a

CONF = cfg.CONF


def is_internal_db_enabled():
    return 'internal-db' in jb_manager.JOB_BINARIES.get_job_binaries()


def check_job_binary_internal(data, **kwargs):
    if not is_internal_db_enabled():
        raise e.BadJobBinaryInternalException(
            _("Sahara internal db is disabled for storing job binaries."))
    if not (isinstance(data, six.binary_type) and len(data) > 0):
        raise e.BadJobBinaryInternalException()
    if "name" in kwargs:
        name = kwargs["name"]
        if not a.validate_name_format(name):
            raise e.BadJobBinaryInternalException(_("%s is not a valid name")
                                                  % name)



# Copyright (c) 2016 Hewlett Packard Enterprise Development Corporation, LP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_middleware import cors


def set_config_defaults():
    """This method updates all configuration default values."""
    set_cors_middleware_defaults()


def set_cors_middleware_defaults():
    """Update default configuration options for oslo.middleware."""

    cors.set_defaults(
        allow_headers=['X-Auth-Token',
                       'X-Identity-Status',
                       'X-Roles',
                       'X-Service-Catalog',
                       'X-User-Id',
                       'X-Tenant-Id',
                       'X-OpenStack-Request-ID'],
        expose_headers=['X-Auth-Token',
                        'X-Subject-Token',
                        'X-Service-Token',
                        'X-OpenStack-Request-ID'],
        allow_methods=['GET',
                       'PUT',
                       'POST',
                       'DELETE',
                       'PATCH']
    )



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import six
import testtools

from sahara import exceptions as ex
from sahara.plugins.fake import plugin
from sahara.service.api import v10 as api
import sahara.service.validation as v
from sahara.service.validations import clusters_scaling as c_s
from sahara.service.validations import clusters_schema as c_schema
from sahara.tests.unit.service.validation import utils as u
from sahara.tests.unit import testutils as tu
from sahara.utils import cluster as c_u


def _get_plugin(plugin_name):
    if plugin_name == 'fake':
        return plugin.FakePluginProvider()
    return None


class TestScalingValidation(u.ValidationTestCase):
    def setUp(self):
        super(TestScalingValidation, self).setUp()
        api.plugin_base.setup_plugins()
        self._create_object_fun = mock.Mock()
        self.duplicates_detected = ("Duplicates in node group names are"
                                    " detected: ['a']")
        self.setup_context(tenant_id='tenant1')

    @mock.patch('sahara.service.api.v10.get_cluster')
    @mock.patch('sahara.plugins.base.PluginManager.get_plugin')
    def _assert_check_scaling(self,
                              get_plugin_p=None,
                              get_cluster_p=None,
                              data=None, cluster=None,
                              expected_message=None,
                              expected_exception=ex.InvalidReferenceException):

        get_cluster_p.return_value = cluster
        get_plugin_p.side_effect = _get_plugin

        with testtools.ExpectedException(expected_exception):
            try:
                c_s.check_cluster_scaling(data, cluster.id)
            except expected_exception as e:
                message = six.text_type(e).split('\n')[0]
                self.assertEqual(expected_message, message)
                raise e

    @mock.patch('sahara.service.api.v10.get_node_group_template')
    @mock.patch('sahara.utils.openstack.nova.client')
    @mock.patch("sahara.service.api.OPS")
    def test_check_cluster_scaling_resize_ng(self, ops, nova_client, get_ngt):
        ops.get_engine_type_and_version.return_value = "direct.1.1"
        ng1 = tu.make_ng_dict('ng', '42', ['namenode'], 1,
                              node_group_template_id='aaa')
        ng2 = tu.make_ng_dict('ng', '42', ['namenode'], 1,
                              resource=True)
        cluster = tu.create_cluster("cluster1", "tenant1", "fake", "0.1",
                                    [ng1],
                                    status=c_u.CLUSTER_STATUS_VALIDATING,
                                    id='12321')

        self._assert_check_scaling(
            data={}, cluster=cluster,
            expected_message="Cluster cannot be scaled "
                             "not in '" + c_u.CLUSTER_STATUS_ACTIVE +
            "' status. Cluster status: " +
            c_u.CLUSTER_STATUS_VALIDATING)

        cluster = tu.create_cluster("cluster1", "tenant1", "fake", "0.1",
                                    [ng1], status=c_u.CLUSTER_STATUS_ACTIVE,
                                    id='12321')
        data = {
            'resize_node_groups': [
                {
                    'name': 'a',
                    'count': 2
                }
            ],
        }
        self._assert_check_scaling(
            data=data, cluster=cluster,
            expected_message="Cluster doesn't contain "
                             "node group with name 'a'")
        data.update({'resize_node_groups': [
            {
                'name': 'a',
                'count': 2
            },
            {
                'name': 'a',
                'count': 2
            }
        ]})
        self._assert_check_scaling(
            data=data, cluster=cluster,
            expected_message=self.duplicates_detected,
            expected_exception=ex.InvalidDataException)

        data = {
            'resize_node_groups': [
                {
                    'name': 'ng',
                    'count': 2
                }
            ],
        }

        client = mock.Mock()
        nova_client.return_value = client
        client.flavors.list.return_value = []
        get_ngt.return_value = ng2

        self._assert_check_scaling(
            data=data, cluster=cluster,
            expected_message="Requested flavor '42' not found",
            expected_exception=ex.NotFoundException)

    @mock.patch("sahara.service.api.OPS")
    def test_check_cluster_scaling_add_ng(self, ops):
        ops.get_engine_type_and_version.return_value = "direct.1.1"
        ng1 = tu.make_ng_dict('ng', '42', ['namenode'], 1)
        cluster = tu.create_cluster("test-cluster", "tenant1", "fake",
                                    "0.1",
                                    [ng1], status=c_u.CLUSTER_STATUS_ACTIVE,
                                    id='12321')
        data = {
            'add_node_groups': [
                {
                    'name': 'a',
                    'flavor_id': '42',
                    'node_processes': ['namenode']
                },
                {
                    'name': 'a',
                    'flavor_id': '42',
                    'node_processes': ['namenode']
                }
            ]
        }
        self._assert_check_scaling(
            data=data, cluster=cluster,
            expected_message=self.duplicates_detected,
            expected_exception=ex.InvalidDataException)
        data = {
            'add_node_groups': [
                {
                    'name': 'ng',
                    'flavor_id': '42',
                    'node_processes': ['namenode']
                },
            ]
        }
        self._assert_check_scaling(
            data=data, cluster=cluster,
            expected_message="Can't add new nodegroup. "
                             "Cluster already has nodegroup "
                             "with name 'ng'")

        data = {
            'add_node_groups': [
                {
                    'name': 'very-very-very-very-very-very-long-ng-name',
                    'flavor_id': '42',
                    'node_processes': ['namenode'],
                    'count': 10
                },
            ]
        }
        patchers = u.start_patch()
        self._assert_check_scaling(
            data=data, cluster=cluster,
            expected_message="Composite hostname test-cluster-very-"
                             "very-very-very-very-very-long-ng-name-"
                             "010.novalocal in provisioned cluster exceeds "
                             "maximum limit 64 characters",
            expected_exception=ex.InvalidDataException)
        u.stop_patch(patchers)

    @mock.patch("sahara.utils.api.request_data")
    @mock.patch("sahara.utils.api.bad_request")
    def _assert_cluster_scaling_validation(self,
                                           bad_req=None,
                                           req_data=None,
                                           data=None,
                                           bad_req_i=None):
        m_func = mock.Mock()
        m_func.__name__ = "m_func"
        req_data.return_value = data
        v.validate(c_schema.CLUSTER_SCALING_SCHEMA,
                   self._create_object_fun)(m_func)(data=data,
                                                    cluster_id='42')

        self.assertEqual(1, req_data.call_count)
        self._assert_calls(bad_req, bad_req_i)

    @mock.patch("sahara.utils.api.request_data")
    @mock.patch("sahara.utils.api.bad_request")
    def _assert_cluster_scaling_validation_v2(self,
                                              bad_req=None,
                                              req_data=None,
                                              data=None,
                                              bad_req_i=None):
        m_func = mock.Mock()
        m_func.__name__ = "m_func"
        req_data.return_value = data
        v.validate(c_schema.CLUSTER_SCALING_SCHEMA_V2,
                   self._create_object_fun)(m_func)(data=data,
                                                    cluster_id='42')

        self.assertEqual(1, req_data.call_count)
        self._assert_calls(bad_req, bad_req_i)

    @mock.patch("sahara.service.api.OPS")
    def test_cluster_scaling_scheme_v_resize_ng(self, ops):
        ops.get_engine_type_and_version.return_value = "direct.1.1"
        self._create_object_fun = mock.Mock()
        data = {
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'{} is not valid under any of the given schemas')
        )
        data = {
            'resize_node_groups': [{}]
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"resize_node_groups[0]: 'name' is a required property")
        )
        data = {
            'resize_node_groups': [
                {
                    'name': 'a'
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"resize_node_groups[0]: 'count' is a required "
                       u"property")
        )
        data = {
            'resize_node_groups': [
                {
                    'name': 'a',
                    'flavor_id': '42',
                    'count': 2
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"resize_node_groups[0]: Additional properties are not "
                       u"allowed ('flavor_id' was unexpected)")
        )

    @mock.patch("sahara.service.api.OPS")
    def test_cluster_scaling_scheme_v_resize_ng_v2(self, ops):
        ops.get_engine_type_and_version.return_value = "direct.1.1"
        self._create_object_fun = mock.Mock()
        data = {
        }
        self._assert_cluster_scaling_validation_v2(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'{} is not valid under any of the given schemas')
        )
        data = {
            'resize_node_groups': [{}]
        }
        self._assert_cluster_scaling_validation_v2(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"resize_node_groups[0]: 'name' is a required property")
        )
        data = {
            'resize_node_groups': [
                {
                    'name': 'a'
                }
            ]
        }
        self._assert_cluster_scaling_validation_v2(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"resize_node_groups[0]: 'count' is a required "
                       u"property")
        )
        data = {
            'resize_node_groups': [
                {
                    'name': 'a',
                    'flavor_id': '42',
                    'instances': ['id1'],
                    'count': 2
                }
            ]
        }
        self._assert_cluster_scaling_validation_v2(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"resize_node_groups[0]: Additional properties are not "
                       u"allowed ('flavor_id' was unexpected)")
        )

    @mock.patch("sahara.service.api.OPS")
    def test_cluster_scaling_validation_add_ng(self, ops):
        ops.get_engine_type_and_version.return_value = "direct.1.1"
        data = {
            'add_node_groups': [
                {
                    "node_group_template_id": "5185a809-6bf7-"
                                              "44ed-9de3-618270550e2c",
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       "add_node_groups[0]: {'node_group_template_id': "
                       "'5185a809-6bf7-44ed-9de3-618270550e2c'} "
                       "is not valid under any of the given schemas")
        )
        data = {
            'add_node_groups': [
                {
                    "node_group_template_id": "5185a809-6bf7-"
                                              "44ed-9de3-618270550e2c",
                    'name': 'a'
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"add_node_groups[0]: {'node_group_template_id': "
                       u"'5185a809-6bf7-44ed-9de3-618270550e2c', "
                       u"'name': 'a'} is not valid under any "
                       u"of the given schemas")
        )

    @mock.patch("sahara.service.api.OPS")
    def test_cluster_scaling_validation_right_schema(self, ops):
        ops.get_engine_type_and_version.return_value = "direct.1.1"
        data = {
            'add_node_groups': [
                {
                    "node_group_template_id": "5185a809-6bf7-"
                                              "44ed-9de3-618270550e2c",
                    'name': 'a',
                    'count': 3
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data
        )
        data = {
            'resize_node_groups': [
                {
                    'name': 'a',
                    'count': 3
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data
        )
        data = {
            'resize_node_groups': [
                {
                    'name': 'a',
                    'count': 3
                }
            ],
            'add_node_groups': [
                {
                    "node_group_template_id": "5185a809-6bf7-"
                                              "44ed-9de3-618270550e2c",
                    'name': 'a',
                    'count': 3
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data
        )

    @mock.patch("sahara.service.api.OPS")
    def test_cluster_scaling_scheme_validation_types(self, ops):
        ops.get_engine_type_and_version.return_value = "direct.1.1"
        data = {
            'resize_node_groups': {},
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"resize_node_groups: {} is not of type 'array'")
        )
        data = {
            'add_node_groups': {}
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"add_node_groups: {} is not of type 'array'")
        )
        data = {
            'resize_node_groups': [],
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'resize_node_groups: [] is too short')
        )

    @mock.patch("sahara.service.api.OPS")
    def test_cluster_scaling_v_right_data(self, ops):
        self.setup_context(tenant_id='t')
        ops.get_engine_type_and_version.return_value = "direct.1.1"
        self._create_object_fun = c_s.check_cluster_scaling

        data = {
            'resize_node_groups': [
                {
                    'name': 'ng',
                    'count': 4
                }
            ],
            'add_node_groups': [
                {
                    'name': 'a',
                    'flavor_id': '42',
                    'node_processes': ['namenode'],
                    'count': 3
                },
            ]
        }
        patchers = u.start_patch()
        self._assert_cluster_scaling_validation(data=data)
        u.stop_patch(patchers)

    @mock.patch("sahara.service.api.OPS")
    def test_check_cluster_scaling_wrong_engine(self, ops):
        ops.get_engine_type_and_version.return_value = "direct.1.1"
        ng1 = tu.make_ng_dict('ng', '42', ['namenode'], 1)
        cluster = tu.create_cluster(
            "cluster1", "tenant1", "fake", "0.1", [ng1],
            status=c_u.CLUSTER_STATUS_ACTIVE, id='12321',
            sahara_info={"infrastructure_engine": "heat.1.1"})

        self._assert_check_scaling(
            data={}, cluster=cluster,
            expected_message="Cluster created with heat.1.1 infrastructure "
                             "engine can't be scaled with direct.1.1 engine")

    @mock.patch("sahara.service.api.OPS")
    def test_check_heat_cluster_scaling_missing_engine(self, ops):
        ops.get_engine_type_and_version.return_value = "heat.1.1"
        ng1 = tu.make_ng_dict('ng', '42', ['namenode'], 1)
        cluster = tu.create_cluster("cluster1", "tenant1", "fake", "0.1",
                                    [ng1], status=c_u.CLUSTER_STATUS_ACTIVE,
                                    id='12321')

        self._assert_check_scaling(
            data={}, cluster=cluster,
            expected_message="Cluster created before Juno release can't be "
                             "scaled with heat.1.1 engine")

    @mock.patch('sahara.utils.openstack.images.image_manager')
    @mock.patch('sahara.utils.openstack.nova.client')
    @mock.patch("sahara.service.api.OPS")
    def test_check_cluster_scaling_missing_resource(self, ops,
                                                    m_nova, m_image):
        ops.get_engine_type_and_version.return_value = "heat.1.1"
        ng1 = tu.make_ng_dict('ng', '42', ['namenode'], 1)

        nova = mock.Mock()
        m_nova.return_value = nova
        nova.keypairs.get.side_effect = u._get_keypair
        cluster = tu.create_cluster(
            "cluster1", "tenant1", "fake", "0.1", [ng1],
            status=c_u.CLUSTER_STATUS_ACTIVE,
            sahara_info={"infrastructure_engine": "heat.1.1"},
            id='12321', user_keypair_id='keypair')
        self._assert_check_scaling(
            data={}, cluster=cluster,
            expected_exception=ex.NotFoundException,
            expected_message="Requested keypair 'keypair' not found")

        image = mock.Mock()
        m_image.return_value = image
        image.list_registered.return_value = [mock.Mock(id='image1'),
                                              mock.Mock(id='image2')]
        cluster = tu.create_cluster(
            "cluster1", "tenant1", "fake", "0.1", [ng1],
            status=c_u.CLUSTER_STATUS_ACTIVE,
            sahara_info={"infrastructure_engine": "heat.1.1"},
            id='12321', default_image_id='image_id',
            user_keypair_id='test_keypair')
        self._assert_check_scaling(
            data={}, cluster=cluster,
            expected_message="Requested image 'image_id' is not registered")
